"""Focused example: nested injected wrappers share one active resolver."""

from __future__ import annotations

from dataclasses import dataclass

from diwire import Container, Injected, Lifetime, Scope, resolver_context


class RequestDependency:
    pass


@dataclass(slots=True)
class InnerService:
    dependency: RequestDependency


@dataclass(slots=True)
class OuterService:
    inner: InnerService
    dependency: RequestDependency


def main() -> None:
    container = Container()
    container.add(
        RequestDependency,
        provides=RequestDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    @resolver_context.inject
    def build_inner(dependency: Injected[RequestDependency]) -> InnerService:
        return InnerService(dependency=dependency)

    @resolver_context.inject
    def build_outer(
        inner: Injected[InnerService],
        dependency: Injected[RequestDependency],
    ) -> OuterService:
        return OuterService(inner=inner, dependency=dependency)

    container.add_factory(
        build_inner,
        provides=InnerService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_factory(
        build_outer,
        provides=OuterService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope() as request_scope:
        resolved = request_scope.resolve(OuterService)

    print(
        f"nested_scope_identity={resolved.inner.dependency is resolved.dependency}",
    )  # => nested_scope_identity=True


if __name__ == "__main__":
    main()



"""Focused example: nested scope work can use ContextVar override/reset tokens."""

from __future__ import annotations

from contextvars import ContextVar

from diwire import Container, Lifetime, Scope

request_id_var: ContextVar[int] = ContextVar("request_id", default=-1)
tenant_var: ContextVar[str] = ContextVar("tenant", default="unset")


def read_request_id() -> int:
    return request_id_var.get()


def read_tenant() -> str:
    return tenant_var.get()


def main() -> None:
    container = Container()
    container.add_factory(
        read_request_id,
        provides=int,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add_factory(
        read_tenant,
        provides=str,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope(Scope.REQUEST) as request_scope:
        request_id_token = request_id_var.set(1)
        tenant_token = tenant_var.set("parent")
        try:
            with request_scope.enter_scope(Scope.ACTION) as action_scope:
                inherited_value = action_scope.resolve(int)

                override_token = request_id_var.set(2)
                try:
                    with action_scope.enter_scope(Scope.STEP) as step_scope:
                        overridden_value = step_scope.resolve(int)
                        inherited_parent_key = step_scope.resolve(str)
                finally:
                    request_id_var.reset(override_token)
        finally:
            tenant_var.reset(tenant_token)
            request_id_var.reset(request_id_token)

    print(f"action_inherits_parent={inherited_value}")  # => action_inherits_parent=1
    print(f"step_overrides_parent={overridden_value}")  # => step_overrides_parent=2
    print(
        f"step_inherits_other_parent_key={inherited_parent_key}"
    )  # => step_inherits_other_parent_key=parent


if __name__ == "__main__":
    main()



from enum import Enum


class MissingPolicy(str, Enum):
    """Policy for handling unresolved registrations."""

    ERROR = "error"
    """Raise an error when a requested type is not registered."""

    REGISTER_ROOT = "register_root"
    """Register only the unresolved root type and do not recurse into dependencies."""

    REGISTER_RECURSIVE = "register_recursive"
    """Register the unresolved type and recursively register its dependencies."""


class DependencyRegistrationPolicy(str, Enum):
    """Policy for handling unresolved dependency types."""

    IGNORE = "ignore"
    """Do not auto-register unresolved dependencies."""

    REGISTER_RECURSIVE = "register_recursive"
    """Recursively register unresolved dependencies."""



from __future__ import annotations

import inspect
from types import SimpleNamespace
from typing import Annotated, Any, cast

import pytest

from diwire import Container, Lifetime, Provider, Scope
from diwire._internal.providers import ProviderDependency, ProviderSpec, ProvidersRegistrations
from diwire._internal.resolvers.assembly import planner as planner_module
from diwire._internal.resolvers.assembly.planner import ResolverGenerationPlanner
from diwire._internal.scope import BaseScope, BaseScopes
from diwire.exceptions import DIWireDependencyNotRegisteredError, DIWireInvalidProviderSpecError


def _planner() -> ResolverGenerationPlanner:
    container = Container()
    container.add_instance(1, provides=int)
    return ResolverGenerationPlanner(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
    )


def test_validate_scope_metadata_errors() -> None:
    with pytest.raises(DIWireInvalidProviderSpecError, match="expected BaseScope"):
        planner_module._validate_resolver_assembly_scope(cast("Any", object()))

    valid_scope = Scope.APP

    monkey = cast("Any", valid_scope)
    original_name = monkey.scope_name
    original_level = monkey.level
    try:
        monkey.scope_name = None
        with pytest.raises(DIWireInvalidProviderSpecError, match="missing 'scope_name'"):
            planner_module._validate_resolver_assembly_scope(valid_scope)

        monkey.scope_name = 1
        with pytest.raises(DIWireInvalidProviderSpecError, match="must be str"):
            planner_module._validate_resolver_assembly_scope(valid_scope)

        monkey.scope_name = "bad.name"
        with pytest.raises(DIWireInvalidProviderSpecError, match="not a valid identifier"):
            planner_module._validate_resolver_assembly_scope(valid_scope)

        monkey.scope_name = "for"
        with pytest.raises(DIWireInvalidProviderSpecError, match="Python keyword"):
            planner_module._validate_resolver_assembly_scope(valid_scope)

        monkey.scope_name = "APP"
        monkey.level = None
        with pytest.raises(DIWireInvalidProviderSpecError, match="missing 'level'"):
            planner_module._validate_resolver_assembly_scope_level(scope=valid_scope)

        monkey.level = "1"
        with pytest.raises(DIWireInvalidProviderSpecError, match="must be int"):
            planner_module._validate_resolver_assembly_scope_level(scope=valid_scope)
    finally:
        monkey.scope_name = original_name
        monkey.level = original_level


def test_validate_managed_scopes_owner_instantiation_error() -> None:
    class _BrokenOwner(BaseScopes):
        def __init__(self) -> None:
            msg = "boom"
            raise RuntimeError(msg)

    root_scope = BaseScope(1)
    cast("Any", root_scope).owner = _BrokenOwner
    with pytest.raises(DIWireInvalidProviderSpecError, match="instantiable"):
        planner_module.validate_resolver_assembly_managed_scopes(root_scope=root_scope)


def test_planner_resolve_provider_attribute_and_cache_owner_errors() -> None:
    planner = _planner()
    scope = Scope.APP

    spec_without_provider = ProviderSpec(
        provides=str,
        dependencies=[],
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=False,
        scope=scope,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(DIWireInvalidProviderSpecError, match="does not define a provider"):
        planner._resolve_provider_attribute(spec=spec_without_provider)

    transient_factory_spec = ProviderSpec(
        provides=float,
        factory=lambda: 1.0,
        dependencies=[],
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=False,
        scope=scope,
        lifetime=Lifetime.TRANSIENT,
    )

    with pytest.raises(DIWireInvalidProviderSpecError, match="Cannot resolve cache owner"):
        planner._cache_owner_scope_level(spec=transient_factory_spec, is_cached=True)


def test_planner_invalid_keyword_argument_name_error_branch() -> None:
    planner = _planner()
    fake_parameter = SimpleNamespace(kind=inspect.Parameter.KEYWORD_ONLY, name="bad-name")
    fake_dependency = SimpleNamespace(parameter=fake_parameter)

    with pytest.raises(DIWireInvalidProviderSpecError, match="not a valid identifier"):
        planner._format_dependency_argument_for_expression(
            dependency=cast("ProviderDependency", fake_dependency),
            expression="value",
        )


def test_planner_dependency_order_without_callable_returns_false() -> None:
    planner = _planner()
    dependency = ProviderDependency(
        provides=int,
        parameter=inspect.Parameter(
            name="value",
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ),
    )
    spec = ProviderSpec(
        provides=bytes,
        instance=b"x",
        dependencies=[dependency],
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=False,
        scope=Scope.APP,
        lifetime=Lifetime.SCOPED,
    )

    assert planner._dependency_order_is_signature_order(spec=spec) is False


def test_planner_max_required_scope_cycle_guard_error() -> None:
    planner = _planner()
    spec = ProviderSpec(
        provides=tuple,
        instance=(),
        dependencies=[],
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=False,
        scope=Scope.APP,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(DIWireInvalidProviderSpecError, match="Circular dependency detected"):
        planner._resolve_max_required_scope_level(
            slot=spec.slot,
            by_slot={spec.slot: spec},
            max_scope_level_by_slot={},
            in_progress={spec.slot},
        )


def test_planner_missing_dependency_raises_from_provider_plan() -> None:
    class _Missing:
        pass

    class _Consumer:
        def __init__(self, dependency: _Missing) -> None:
            self.dependency = dependency

    registrations = ProvidersRegistrations()
    dependency = ProviderDependency(
        provides=_Missing,
        parameter=inspect.Parameter(
            name="dependency",
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ),
    )
    registrations.add(
        ProviderSpec(
            provides=_Consumer,
            concrete_type=_Consumer,
            dependencies=[dependency],
            is_async=False,
            is_any_dependency_async=False,
            needs_cleanup=False,
            scope=Scope.APP,
            lifetime=Lifetime.SCOPED,
        ),
    )

    with pytest.raises(DIWireDependencyNotRegisteredError, match="is not registered"):
        ResolverGenerationPlanner(
            root_scope=Scope.APP,
            registrations=registrations,
        ).build()


def test_planner_dependency_order_out_of_signature_order_returns_false() -> None:
    planner = _planner()

    def _factory(first: int, second: int) -> int:
        return first + second

    dependencies = [
        ProviderDependency(
            provides=int,
            parameter=inspect.Parameter(
                name="second",
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            ),
        ),
        ProviderDependency(
            provides=int,
            parameter=inspect.Parameter(
                name="first",
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            ),
        ),
    ]
    spec = ProviderSpec(
        provides=int,
        factory=_factory,
        dependencies=dependencies,
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=False,
        scope=Scope.APP,
        lifetime=Lifetime.SCOPED,
    )

    assert planner._dependency_order_is_signature_order(spec=spec) is False


def test_planner_provider_callable_for_generator_and_context_manager() -> None:
    planner = _planner()

    def _generator() -> Any:
        yield 1

    class _ContextManager:
        def __enter__(self) -> int:
            return 1

        def __exit__(self, *_args: object) -> None:
            return None

    def _context_manager() -> _ContextManager:
        return _ContextManager()

    generator_spec = ProviderSpec(
        provides=list,
        generator=_generator,
        dependencies=[],
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=True,
        scope=Scope.APP,
        lifetime=Lifetime.SCOPED,
    )
    context_manager_spec = ProviderSpec(
        provides=dict,
        context_manager=_context_manager,
        dependencies=[],
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=True,
        scope=Scope.APP,
        lifetime=Lifetime.SCOPED,
    )

    assert planner._provider_callable_for_signature(spec=generator_spec) is _generator
    assert planner._provider_callable_for_signature(spec=context_manager_spec) is _context_manager


def test_plan_provider_dependency_missing_required_uses_dependency_spec_error() -> None:
    planner = _planner()
    registration_spec = planner._registrations.get_by_type(int)
    dependency = ProviderDependency(
        provides=str,
        parameter=inspect.Parameter(
            name="dependency",
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ),
    )

    with pytest.raises(DIWireDependencyNotRegisteredError, match="is not registered"):
        planner._plan_provider_dependency(
            spec=registration_spec,
            dependency_index=0,
            dependency=dependency,
            dependency_key=str,
            optional=False,
        )


def test_plan_provider_dependency_uses_normalized_key_fallback() -> None:
    planner = _planner()
    registration_spec = planner._registrations.get_by_type(int)
    dependency_key = Annotated[int, "meta"]
    dependency = ProviderDependency(
        provides=dependency_key,
        parameter=inspect.Parameter(
            name="dependency",
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ),
    )

    _plan, dependency_slot, _requires_async, _sync_arg, _async_arg = (
        planner._plan_provider_dependency(
            spec=registration_spec,
            dependency_index=0,
            dependency=dependency,
            dependency_key=dependency_key,
            optional=False,
        )
    )

    assert dependency_slot == registration_spec.slot


def test_plan_provider_handle_dependency_uses_normalized_inner_key_fallback() -> None:
    planner = _planner()
    registration_spec = planner._registrations.get_by_type(int)
    dependency_key = Provider[Annotated[int, "meta"]]
    dependency = ProviderDependency(
        provides=dependency_key,
        parameter=inspect.Parameter(
            name="provider",
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ),
    )

    plan, _dependency_slot, _requires_async, _sync_arg, _async_arg = (
        planner._plan_provider_handle_dependency(
            spec=registration_spec,
            dependency_index=0,
            dependency=dependency,
            dependency_key=dependency_key,
        )
    )

    assert plan.provider_inner_slot == registration_spec.slot


def test_dependency_slots_for_graph_uses_normalized_key_fallback() -> None:
    planner = _planner()
    dependency_key = Annotated[int, "meta"]
    dependency = ProviderDependency(
        provides=dependency_key,
        parameter=inspect.Parameter(
            name="dependency",
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ),
    )

    slots = planner._dependency_slots_for_graph(
        dependency=dependency,
        requiring_provider=int,
    )

    assert slots == (planner._registrations.get_by_type(int).slot,)


def test_dependency_spec_or_error_uses_normalized_key_fallback() -> None:
    planner = _planner()

    dependency_spec = planner._dependency_spec_or_error(
        dependency_provides=Annotated[int, "meta"],
        requiring_provider=int,
    )

    assert dependency_spec.provides is int


def test_build_all_slots_by_key_includes_non_component_annotated_registrations() -> None:
    container = Container()
    container.add_instance(1, provides=Annotated[int, "meta"])
    container.add_instance(2, provides=int, component="primary")
    planner = ResolverGenerationPlanner(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
    )

    slots_by_key = planner._build_all_slots_by_key()

    assert int in slots_by_key
    assert len(slots_by_key[int]) == 2



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, make_dataclass
from typing import Any

import rodi
from dishka import Provider
from wireup import injectable

from diwire import Container, Lifetime, Scope
from tests.benchmarks.dishka_helpers import DishkaBenchmarkScope, make_dishka_benchmark_container
from tests.benchmarks.helpers import make_diwire_benchmark_container, run_benchmark
from tests.benchmarks.wireup_helpers import make_wireup_benchmark_container

# Wireup container compilation time becomes pathological with larger generated graphs.
# Keep this scenario large enough to exercise scoped graph behavior while remaining runnable across frameworks.
GRID_WIDTH = 6
GRID_HEIGHT = 6


@dataclass(frozen=True, slots=True)
class _GridGraph:
    tops: tuple[type[Any], ...]
    layers: tuple[tuple[type[Any], ...], ...]
    bottom: type[Any]


def _make_node_class(name: str, deps: Sequence[type[Any]]) -> type[Any]:
    return make_dataclass(
        name,
        [(f"d_{index}", dep) for index, dep in enumerate(deps)],
        slots=True,
        namespace={"__module__": __name__},
    )


def _build_scoped_grid_graph(*, width: int, height: int) -> _GridGraph:
    tops = tuple(_make_node_class(f"_Top_{index}", ()) for index in range(width))
    layers: list[tuple[type[Any], ...]] = []

    prev_classes: tuple[type[Any], ...] = tops
    for level in range(height):
        layer = tuple(
            _make_node_class(f"_Middle_{level}_{index}", prev_classes) for index in range(width)
        )
        layers.append(layer)
        prev_classes = layer

    bottom = _make_node_class("_Bottom", prev_classes)
    return _GridGraph(
        tops=tops,
        layers=tuple(layers),
        bottom=bottom,
    )


def _build_scoped_grid_diwire_container(*, graph: _GridGraph) -> Container:
    container = make_diwire_benchmark_container()

    for cls in graph.tops:
        container.add(cls, lifetime=Lifetime.SCOPED, scope=Scope.APP)
    for layer in graph.layers:
        for cls in layer:
            container.add(cls, lifetime=Lifetime.SCOPED, scope=Scope.REQUEST)
    container.add(graph.bottom, lifetime=Lifetime.SCOPED, scope=Scope.REQUEST)
    container.compile()
    return container


def _build_scoped_grid_dishka_container(*, graph: _GridGraph) -> Any:
    provider = Provider(scope=DishkaBenchmarkScope.APP)

    for cls in graph.tops:
        provider.provide(cls, scope=DishkaBenchmarkScope.APP)
    for layer in graph.layers:
        for cls in layer:
            provider.provide(cls, scope=DishkaBenchmarkScope.REQUEST)
    provider.provide(graph.bottom, scope=DishkaBenchmarkScope.REQUEST)
    return make_dishka_benchmark_container(provider)


def _build_scoped_grid_rodi_container(*, graph: _GridGraph) -> Any:
    container = rodi.Container()

    for cls in graph.tops:
        container.add_singleton(cls)
    for layer in graph.layers:
        for cls in layer:
            container.add_scoped(cls)
    container.add_scoped(graph.bottom)
    return container.build_provider()


def _wireup_injectable(cls: type[Any], *, lifetime: str) -> type[Any]:
    return injectable(lifetime=lifetime)(cls)


def _build_scoped_grid_wireup_container(*, graph: _GridGraph) -> Any:
    injectables = [_wireup_injectable(cls, lifetime="singleton") for cls in graph.tops]
    for layer in graph.layers:
        injectables.extend(_wireup_injectable(cls, lifetime="scoped") for cls in layer)
    injectables.append(_wireup_injectable(graph.bottom, lifetime="scoped"))

    return make_wireup_benchmark_container(*injectables)


def _walk_to_top(node: Any, *, height: int) -> Any:
    current = node.d_0
    for _ in range(height):
        current = current.d_0
    return current


def test_benchmark_diwire_resolve_generated_scoped_grid(benchmark: Any) -> None:
    graph = _build_scoped_grid_graph(width=GRID_WIDTH, height=GRID_HEIGHT)
    container = _build_scoped_grid_diwire_container(graph=graph)

    with container.enter_scope(Scope.REQUEST) as first_scope:
        first = first_scope.resolve(graph.bottom)
        second = first_scope.resolve(graph.bottom)
    with container.enter_scope(Scope.REQUEST) as second_scope:
        third = second_scope.resolve(graph.bottom)

    assert first is second
    assert first is not third
    assert first.d_0 is not third.d_0

    first_top = _walk_to_top(first, height=GRID_HEIGHT)
    third_top = _walk_to_top(third, height=GRID_HEIGHT)
    assert first_top is third_top

    def bench_diwire_generated_scoped_grid() -> None:
        with container.enter_scope(Scope.REQUEST) as scope:
            _ = scope.resolve(graph.bottom)

    run_benchmark(benchmark, bench_diwire_generated_scoped_grid, iterations=500)


def test_benchmark_dishka_resolve_generated_scoped_grid(benchmark: Any) -> None:
    graph = _build_scoped_grid_graph(width=GRID_WIDTH, height=GRID_HEIGHT)
    container = _build_scoped_grid_dishka_container(graph=graph)

    with container(scope=DishkaBenchmarkScope.REQUEST) as first_scope:
        first = first_scope.get(graph.bottom)
        second = first_scope.get(graph.bottom)
    with container(scope=DishkaBenchmarkScope.REQUEST) as second_scope:
        third = second_scope.get(graph.bottom)

    assert first is second
    assert first is not third
    assert first.d_0 is not third.d_0

    first_top = _walk_to_top(first, height=GRID_HEIGHT)
    third_top = _walk_to_top(third, height=GRID_HEIGHT)
    assert first_top is third_top

    def bench_dishka_generated_scoped_grid() -> None:
        with container(scope=DishkaBenchmarkScope.REQUEST) as scope:
            _ = scope.get(graph.bottom)

    run_benchmark(benchmark, bench_dishka_generated_scoped_grid, iterations=500)


def test_benchmark_rodi_resolve_generated_scoped_grid(benchmark: Any) -> None:
    graph = _build_scoped_grid_graph(width=GRID_WIDTH, height=GRID_HEIGHT)
    container = _build_scoped_grid_rodi_container(graph=graph)

    with container.create_scope() as first_scope:
        first = first_scope.get(graph.bottom)
        second = first_scope.get(graph.bottom)
    with container.create_scope() as second_scope:
        third = second_scope.get(graph.bottom)

    assert first is second
    assert first is not third
    assert first.d_0 is not third.d_0

    first_top = _walk_to_top(first, height=GRID_HEIGHT)
    third_top = _walk_to_top(third, height=GRID_HEIGHT)
    assert first_top is third_top

    def bench_rodi_generated_scoped_grid() -> None:
        with container.create_scope() as scope:
            _ = scope.get(graph.bottom)

    run_benchmark(benchmark, bench_rodi_generated_scoped_grid, iterations=500)


def test_benchmark_wireup_resolve_generated_scoped_grid(benchmark: Any) -> None:
    graph = _build_scoped_grid_graph(width=GRID_WIDTH, height=GRID_HEIGHT)
    container = _build_scoped_grid_wireup_container(graph=graph)

    with container.enter_scope() as first_scope:
        first = first_scope.get(graph.bottom)
        second = first_scope.get(graph.bottom)
    with container.enter_scope() as second_scope:
        third = second_scope.get(graph.bottom)

    assert first is second
    assert first is not third
    assert first.d_0 is not third.d_0

    first_top = _walk_to_top(first, height=GRID_HEIGHT)
    third_top = _walk_to_top(third, height=GRID_HEIGHT)
    assert first_top is third_top

    def bench_wireup_generated_scoped_grid() -> None:
        with container.enter_scope() as scope:
            _ = scope.get(graph.bottom)

    run_benchmark(benchmark, bench_wireup_generated_scoped_grid, iterations=500)



from __future__ import annotations

from collections.abc import Awaitable, Callable
from contextvars import ContextVar
from typing import TypeAlias, cast

from asgiref.sync import iscoroutinefunction, markcoroutinefunction
from django.http import HttpRequest, HttpResponseBase

from diwire import Container, Lifetime
from diwire.exceptions import DIWireIntegrationError

_request_context: ContextVar[HttpRequest | None] = ContextVar(
    "diwire_django_request_context",
    default=None,
)

_GetResponseSync: TypeAlias = Callable[..., HttpResponseBase]
_GetResponseAsync: TypeAlias = Callable[..., Awaitable[HttpResponseBase]]


def get_request() -> HttpRequest:
    request = _request_context.get()
    if request is None:
        msg = "Request context not available. Ensure RequestContextMiddleware is installed."
        raise DIWireIntegrationError(msg)
    return request


class RequestContextMiddleware:
    sync_capable = True
    async_capable = True

    def __init__(
        self,
        get_response: _GetResponseSync | _GetResponseAsync,
    ) -> None:
        self.get_response = get_response
        self._is_async = iscoroutinefunction(get_response)
        if self._is_async:
            markcoroutinefunction(self)

    def __call__(
        self,
        request: HttpRequest,
    ) -> HttpResponseBase | Awaitable[HttpResponseBase]:
        if self._is_async:
            return self._call_async(request)
        return self._call_sync(request)

    def _call_sync(self, request: HttpRequest) -> HttpResponseBase:
        token = _request_context.set(request)
        try:
            return cast("_GetResponseSync", self.get_response)(request)
        finally:
            _request_context.reset(token)

    async def _call_async(self, request: HttpRequest) -> HttpResponseBase:
        token = _request_context.set(request)
        try:
            return await cast("_GetResponseAsync", self.get_response)(request)
        finally:
            _request_context.reset(token)


def add_request_context(container: Container) -> None:
    container.add_factory(
        get_request,
        provides=HttpRequest,
        lifetime=Lifetime.TRANSIENT,
    )



from __future__ import annotations

from collections.abc import Generator
from typing import Any

from dishka import Provider

from diwire import Lifetime, Scope
from tests.benchmarks.dishka_helpers import DishkaBenchmarkScope, make_dishka_benchmark_container
from tests.benchmarks.helpers import make_diwire_benchmark_container, run_benchmark


class _RequestResource:
    pass


def test_benchmark_diwire_enter_close_scope_resolve_generator_request_try_finally(
    benchmark: Any,
) -> None:
    cleanup_count = 0

    def build_request_resource() -> Generator[_RequestResource, None, None]:
        nonlocal cleanup_count
        resource = _RequestResource()
        try:
            yield resource
        finally:
            cleanup_count += 1

    container = make_diwire_benchmark_container()
    container.add_generator(
        build_request_resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.compile()

    with container.enter_scope(Scope.REQUEST) as first_scope:
        first = first_scope.resolve(_RequestResource)
        second = first_scope.resolve(_RequestResource)
    assert first is second
    assert cleanup_count == 1

    with container.enter_scope(Scope.REQUEST) as second_scope:
        third = second_scope.resolve(_RequestResource)
    assert first is not third
    assert cleanup_count == 2

    def bench_diwire_enter_close_scope_resolve_generator_request_try_finally() -> None:
        with container.enter_scope(Scope.REQUEST) as scope:
            _ = scope.resolve(_RequestResource)

    run_benchmark(
        benchmark,
        bench_diwire_enter_close_scope_resolve_generator_request_try_finally,
        iterations=25_000,
    )


def test_benchmark_dishka_enter_close_scope_resolve_generator_request_try_finally(
    benchmark: Any,
) -> None:
    cleanup_count = 0

    def build_request_resource() -> Generator[_RequestResource, None, None]:
        nonlocal cleanup_count
        resource = _RequestResource()
        try:
            yield resource
        finally:
            cleanup_count += 1

    provider = Provider(scope=DishkaBenchmarkScope.APP)
    provider.provide(
        build_request_resource,
        scope=DishkaBenchmarkScope.REQUEST,
    )
    container = make_dishka_benchmark_container(provider)

    with container(scope=DishkaBenchmarkScope.REQUEST) as first_scope:
        first = first_scope.get(_RequestResource)
        second = first_scope.get(_RequestResource)
    assert first is second
    assert cleanup_count == 1

    with container(scope=DishkaBenchmarkScope.REQUEST) as second_scope:
        third = second_scope.get(_RequestResource)
    assert first is not third
    assert cleanup_count == 2

    def bench_dishka_enter_close_scope_resolve_generator_request_try_finally() -> None:
        with container(scope=DishkaBenchmarkScope.REQUEST) as scope:
            _ = scope.get(_RequestResource)

    run_benchmark(
        benchmark,
        bench_dishka_enter_close_scope_resolve_generator_request_try_finally,
        iterations=25_000,
    )



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeAlias

import pytest

from diwire import Container, Injected, Lifetime, Scope, resolver_context

if TYPE_CHECKING:
    from litestar import Litestar
    from litestar.connection import Request, WebSocket
    from litestar.datastructures.state import State
    from litestar.testing import TestClient

LitestarRequest: TypeAlias = "Request[object, object, State]"
LitestarWebSocket: TypeAlias = "WebSocket[object, object, State]"


@dataclass
class _RequestPathService:
    request: LitestarRequest

    def path(self) -> str:
        return self.request.url.path


@dataclass
class _WebSocketPathService:
    websocket: LitestarWebSocket

    def path(self) -> str:
        return self.websocket.url.path


@pytest.fixture()
def app() -> Litestar:
    pytest.importorskip("litestar")

    from litestar import Litestar, get, websocket
    from litestar.connection import Request, WebSocket
    from litestar.datastructures.state import State

    from diwire.integrations.litestar import RequestContextMiddleware, add_request_context

    # diwire resolves dependencies via runtime type hints; populate module globals so
    # forward references like `Request[object, object, State]` can be evaluated.
    globals()["Request"] = Request
    globals()["WebSocket"] = WebSocket
    globals()["State"] = State

    container = Container()
    add_request_context(container)
    container.add(
        _RequestPathService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _WebSocketPathService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    @get("/request/direct")
    @resolver_context.inject(scope=Scope.REQUEST)
    async def request_direct(
        request: LitestarRequest,
        resolved_request: Injected[LitestarRequest],
    ) -> dict[str, str]:
        _ = request
        return {"path": resolved_request.url.path}

    @get("/request/service")
    @resolver_context.inject(scope=Scope.REQUEST)
    async def request_service(service: Injected[_RequestPathService]) -> dict[str, str]:
        return {"path": service.path()}

    @websocket("/websocket/direct")
    @resolver_context.inject(scope=Scope.REQUEST)
    async def websocket_direct(
        socket: LitestarWebSocket,
        resolved_websocket: Injected[LitestarWebSocket],
    ) -> None:
        await socket.accept()
        await socket.send_json({"path": resolved_websocket.url.path})
        await socket.close()

    @websocket("/websocket/service")
    @resolver_context.inject(scope=Scope.REQUEST)
    async def websocket_service(
        socket: LitestarWebSocket,
        service: Injected[_WebSocketPathService],
    ) -> None:
        await socket.accept()
        await socket.send_json({"path": service.path()})
        await socket.close()

    return Litestar(
        route_handlers=[
            request_direct,
            request_service,
            websocket_direct,
            websocket_service,
        ],
        middleware=[RequestContextMiddleware()],
    )


@pytest.fixture()
def client(app: Litestar) -> Iterator[TestClient[Litestar]]:
    pytest.importorskip("litestar")

    from litestar.testing import TestClient

    with TestClient(app) as test_client:
        yield test_client


def test_request_resolve_for_http_endpoint(client: TestClient[Litestar]) -> None:
    response = client.get("/request/direct")
    assert response.status_code == 200
    assert response.json() == {"path": "/request/direct"}


def test_request_resolve_in_service_for_http_endpoint(client: TestClient[Litestar]) -> None:
    response = client.get("/request/service")
    assert response.status_code == 200
    assert response.json() == {"path": "/request/service"}


def test_websocket_resolve_for_websocket_endpoint(client: TestClient[Litestar]) -> None:
    with client.websocket_connect("/websocket/direct") as websocket:
        payload = websocket.receive_json()
    assert payload == {"path": "/websocket/direct"}


def test_websocket_resolve_in_service_for_websocket_endpoint(client: TestClient[Litestar]) -> None:
    with client.websocket_connect("/websocket/service") as websocket:
        payload = websocket.receive_json()
    assert payload == {"path": "/websocket/service"}



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, TypeVar, cast

from dishka import Provider

from diwire import Lifetime, Scope
from tests.benchmarks.dishka_helpers import DishkaBenchmarkScope, make_dishka_benchmark_container
from tests.benchmarks.helpers import make_diwire_benchmark_container, run_benchmark

T = TypeVar("T")


class _Repo(Generic[T]):
    pass


@dataclass
class _OpenGenericRepo(_Repo[T]):
    dependency_type: type[T]


class _ClosedIntRepo(_Repo[int]):
    pass


def test_benchmark_diwire_resolve_scoped_with_registered_open_closed_generics(
    benchmark: Any,
) -> None:
    container = make_diwire_benchmark_container()
    container.add(
        _OpenGenericRepo,
        provides=_Repo,
        lifetime=Lifetime.SCOPED,
        scope=Scope.REQUEST,
    )
    container.add(
        _ClosedIntRepo,
        provides=_Repo[int],
        lifetime=Lifetime.SCOPED,
        scope=Scope.REQUEST,
    )
    container.compile()

    with container.enter_scope(Scope.REQUEST) as first_scope:
        first_int = cast("Any", first_scope.resolve(_Repo[int]))
        second_int = cast("Any", first_scope.resolve(_Repo[int]))
        first_str = cast("Any", first_scope.resolve(_Repo[str]))
    with container.enter_scope(Scope.REQUEST) as second_scope:
        third_int = cast("Any", second_scope.resolve(_Repo[int]))
        second_str = cast("Any", second_scope.resolve(_Repo[str]))
    assert isinstance(first_int, _ClosedIntRepo)
    assert isinstance(second_int, _ClosedIntRepo)
    assert isinstance(third_int, _ClosedIntRepo)
    assert isinstance(first_str, _OpenGenericRepo)
    assert isinstance(second_str, _OpenGenericRepo)
    assert first_int is second_int
    assert first_int is not third_int
    assert first_str is not second_str
    assert first_str.dependency_type is str
    assert second_str.dependency_type is str

    repo_int = _Repo[int]
    with container.enter_scope(Scope.REQUEST) as bench_scope:
        _ = bench_scope.resolve(repo_int)

        def bench_diwire_resolve_scoped_with_registered_open_closed_generics() -> None:
            _ = bench_scope.resolve(repo_int)

        run_benchmark(benchmark, bench_diwire_resolve_scoped_with_registered_open_closed_generics)


def test_benchmark_dishka_resolve_scoped_with_registered_open_closed_generics(
    benchmark: Any,
) -> None:
    provider = Provider(scope=DishkaBenchmarkScope.APP)
    provider.provide(_OpenGenericRepo, provides=_Repo, scope=DishkaBenchmarkScope.REQUEST)
    provider.provide(_ClosedIntRepo, provides=_Repo[int], scope=DishkaBenchmarkScope.REQUEST)
    container = make_dishka_benchmark_container(provider)

    with container(scope=DishkaBenchmarkScope.REQUEST) as first_scope:
        first_int = cast("Any", first_scope.get(_Repo[int]))
        second_int = cast("Any", first_scope.get(_Repo[int]))
        first_str = cast("Any", first_scope.get(_Repo[str]))
    with container(scope=DishkaBenchmarkScope.REQUEST) as second_scope:
        third_int = cast("Any", second_scope.get(_Repo[int]))
        second_str = cast("Any", second_scope.get(_Repo[str]))
    assert isinstance(first_int, _ClosedIntRepo)
    assert isinstance(second_int, _ClosedIntRepo)
    assert isinstance(third_int, _ClosedIntRepo)
    assert isinstance(first_str, _OpenGenericRepo)
    assert isinstance(second_str, _OpenGenericRepo)
    assert first_int is second_int
    assert first_int is not third_int
    assert first_str is not second_str
    assert first_str.dependency_type is str
    assert second_str.dependency_type is str

    repo_int = _Repo[int]
    with container(scope=DishkaBenchmarkScope.REQUEST) as bench_scope:
        _ = bench_scope.get(repo_int)

        def bench_dishka_resolve_scoped_with_registered_open_closed_generics() -> None:
            _ = bench_scope.get(repo_int)

        run_benchmark(benchmark, bench_dishka_resolve_scoped_with_registered_open_closed_generics)



from __future__ import annotations

from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from json import loads
from types import TracebackType
from typing import TYPE_CHECKING, cast

import pytest
from typing_extensions import Self

pytest.importorskip("django")

from django.http import HttpRequest, HttpResponseBase, JsonResponse

from diwire import Container, Injected, Lifetime, Scope, resolver_context
from tests._django_setup import ensure_django_setup

if TYPE_CHECKING:
    from diwire.integrations.django import RequestContextMiddleware

ensure_django_setup()

_cleanup_paths: list[str] = []


def _latest_cleanup_path() -> str | None:
    if not _cleanup_paths:
        return None
    return _cleanup_paths[-1]


def _build_request(path: str) -> HttpRequest:
    request = HttpRequest()
    request.path = path
    request.path_info = path
    return request


def _json_body(response: HttpResponseBase) -> dict[str, str]:
    return cast("dict[str, str]", loads(response.content))


@dataclass
class _RequestPathService:
    request: HttpRequest

    def path(self) -> str:
        return self.request.path


@dataclass
class _CMService:
    request: HttpRequest

    def work(self) -> str:
        return "CMService working"

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        _cleanup_paths.append(self.request.path)


@pytest.fixture()
def handlers() -> dict[str, Callable[[HttpRequest], HttpResponseBase] | RequestContextMiddleware]:
    from diwire.integrations.django import RequestContextMiddleware, add_request_context

    _cleanup_paths.clear()

    container = Container()
    add_request_context(container)
    container.add(
        _RequestPathService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_context_manager(_CMService, scope=Scope.REQUEST)

    @resolver_context.inject(scope=Scope.REQUEST)
    def request_direct(
        request: HttpRequest,
        resolved_request: Injected[HttpRequest],
    ) -> JsonResponse:
        _ = request
        return JsonResponse({"path": resolved_request.path})

    @resolver_context.inject(scope=Scope.REQUEST)
    def request_service(
        request: HttpRequest,
        service: Injected[_RequestPathService],
    ) -> JsonResponse:
        _ = request
        return JsonResponse({"path": service.path()})

    @resolver_context.inject(scope=Scope.REQUEST)
    def cm_service(
        request: HttpRequest,
        service: Injected[_CMService],
    ) -> JsonResponse:
        _ = request
        return JsonResponse({"message": service.work()})

    @resolver_context.inject(scope=Scope.REQUEST)
    async def request_service_async(
        request: HttpRequest,
        service: Injected[_RequestPathService],
    ) -> JsonResponse:
        _ = request
        return JsonResponse({"path": service.path()})

    return {
        "request_direct": RequestContextMiddleware(request_direct),
        "request_service": RequestContextMiddleware(request_service),
        "cm_service": RequestContextMiddleware(cm_service),
        "request_service_async": RequestContextMiddleware(request_service_async),
    }


def test_request_resolve_for_http_endpoint(
    handlers: dict[str, Callable[[HttpRequest], HttpResponseBase] | RequestContextMiddleware],
) -> None:
    request = _build_request("/request/direct")
    handler = cast("Callable[[HttpRequest], HttpResponseBase]", handlers["request_direct"])
    response = handler(request)
    assert response.status_code == 200
    assert _json_body(response) == {"path": "/request/direct"}


def test_request_resolve_in_service_for_http_endpoint(
    handlers: dict[str, Callable[[HttpRequest], HttpResponseBase] | RequestContextMiddleware],
) -> None:
    request = _build_request("/request/service")
    handler = cast("Callable[[HttpRequest], HttpResponseBase]", handlers["request_service"])
    response = handler(request)
    assert response.status_code == 200
    assert _json_body(response) == {"path": "/request/service"}


def test_context_manager_service_validates_request_scope_cleanup_path(
    handlers: dict[str, Callable[[HttpRequest], HttpResponseBase] | RequestContextMiddleware],
) -> None:
    request = _build_request("/services/cm-service")
    handler = cast("Callable[[HttpRequest], HttpResponseBase]", handlers["cm_service"])
    response = handler(request)
    assert response.status_code == 200
    assert _json_body(response) == {"message": "CMService working"}
    assert _latest_cleanup_path() == "/services/cm-service"


@pytest.mark.asyncio
async def test_request_resolve_in_service_for_async_http_endpoint(
    handlers: dict[str, Callable[[HttpRequest], HttpResponseBase] | RequestContextMiddleware],
) -> None:
    request = _build_request("/request/service/async")
    handler = cast(
        "Callable[[HttpRequest], Awaitable[HttpResponseBase]]",
        handlers["request_service_async"],
    )
    response = await handler(request)
    assert response.status_code == 200
    assert _json_body(response) == {"path": "/request/service/async"}



from __future__ import annotations

from typing import cast

import click
import typer

from diwire import Container, Lifetime
from diwire.exceptions import DIWireIntegrationError


def get_typer_context() -> typer.Context:
    try:
        return cast("typer.Context", click.get_current_context())
    except RuntimeError as error:
        msg = "Typer context not available. Ensure the command runs inside a Typer application."
        raise DIWireIntegrationError(msg) from error


def add_typer_context(container: Container) -> None:
    container.add_factory(
        get_typer_context,
        provides=typer.Context,
        lifetime=Lifetime.TRANSIENT,
    )



from __future__ import annotations

import argparse
import json
import platform
import statistics
import sys
import timeit
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any


@dataclass(frozen=True, slots=True)
class ScenarioResult:
    """A single benchmark row comparing legacy and hybrid dispatch timings."""

    name: str
    before_ns_op: float
    after_ns_op: float
    ratio_after_over_before: float
    before_behavior: str
    after_behavior: str


def _compile_identity_dispatch(keys: tuple[Any, ...]) -> Any:
    lines = ["def resolve(dependency):"]
    for index in range(len(keys)):
        lines.append(f"    if dependency is _dep_{index}:")
        lines.append(f"        return {index}")
    lines.append("    raise KeyError(dependency)")
    source = "\n".join(lines)
    namespace: dict[str, Any] = {f"_dep_{index}": key for index, key in enumerate(keys)}
    exec(source, namespace)  # noqa: S102
    return namespace["resolve"]


def _compile_hybrid_dispatch(
    *,
    identity_keys: tuple[Any, ...],
    equality_keys: tuple[Any, ...],
) -> Any:
    equality_slot_by_key = {
        key: len(identity_keys) + index for index, key in enumerate(equality_keys)
    }
    lines = ["def resolve(dependency):"]
    for index in range(len(identity_keys)):
        lines.append(f"    if dependency is _dep_{index}:")
        lines.append(f"        return {index}")
    if equality_slot_by_key:
        lines.extend(
            [
                "    slot = _dep_eq_slot_by_key.get(dependency, _missing_slot)",
                "    if slot is not _missing_slot:",
                "        return slot",
            ],
        )
    lines.append("    raise KeyError(dependency)")
    source = "\n".join(lines)
    namespace: dict[str, Any] = {
        **{f"_dep_{index}": key for index, key in enumerate(identity_keys)},
        "_dep_eq_slot_by_key": equality_slot_by_key,
        "_missing_slot": object(),
    }
    exec(source, namespace)  # noqa: S102
    return namespace["resolve"]


def _measure_pair_ns_per_op(
    *,
    before_callable: Any,
    after_callable: Any,
    number: int,
    repeat: int,
) -> tuple[float, float]:
    before_timer = timeit.Timer(before_callable)
    after_timer = timeit.Timer(after_callable)
    before_callable()
    after_callable()

    before_runs = [before_timer.timeit(number=number) for _ in range(repeat)]
    after_runs = [after_timer.timeit(number=number) for _ in range(repeat)]
    before_ns = (statistics.median(before_runs) * 1_000_000_000.0) / number
    after_ns = (statistics.median(after_runs) * 1_000_000_000.0) / number
    return before_ns, after_ns


def _class_keys(count: int) -> tuple[type[Any], ...]:
    return tuple(type(f"_DispatchBenchClass{index}", (), {}) for index in range(count))


def _scenario_class_path(
    *,
    provider_count: int,
    position_name: str,
    index: int,
    number: int,
    repeat: int,
) -> ScenarioResult:
    keys = _class_keys(provider_count)
    dependency = keys[index]
    before_dispatch = _compile_identity_dispatch(keys)
    after_dispatch = _compile_hybrid_dispatch(identity_keys=keys, equality_keys=())

    def before_call() -> None:
        _ = before_dispatch(dependency)

    def after_call() -> None:
        _ = after_dispatch(dependency)

    before_ns, after_ns = _measure_pair_ns_per_op(
        before_callable=before_call,
        after_callable=after_call,
        number=number,
        repeat=repeat,
    )
    return ScenarioResult(
        name=f"class_{position_name}_n{provider_count}",
        before_ns_op=before_ns,
        after_ns_op=after_ns,
        ratio_after_over_before=after_ns / before_ns,
        before_behavior="hit",
        after_behavior="hit",
    )


def _scenario_alias_equal_not_identical(
    *,
    alias_name: str,
    alias_factory: Any,
    number: int,
    repeat: int,
) -> ScenarioResult:
    class_keys = _class_keys(32)
    registration_alias = alias_factory()
    lookup_alias = alias_factory()
    if lookup_alias is registration_alias or lookup_alias != registration_alias:
        msg = f"Expected equal but non-identical alias key for {alias_name}."
        raise RuntimeError(msg)

    before_dispatch = _compile_identity_dispatch((*class_keys, registration_alias))
    after_dispatch = _compile_hybrid_dispatch(
        identity_keys=class_keys,
        equality_keys=(registration_alias,),
    )

    def before_call() -> None:
        try:
            before_dispatch(lookup_alias)
        except KeyError:
            return
        msg = "Legacy identity dispatch unexpectedly resolved an equal non-identical alias."
        raise RuntimeError(msg)

    def after_call() -> None:
        _ = after_dispatch(lookup_alias)

    before_ns, after_ns = _measure_pair_ns_per_op(
        before_callable=before_call,
        after_callable=after_call,
        number=number,
        repeat=repeat,
    )
    return ScenarioResult(
        name=f"alias_equal_not_identical_{alias_name}",
        before_ns_op=before_ns,
        after_ns_op=after_ns,
        ratio_after_over_before=after_ns / before_ns,
        before_behavior="miss",
        after_behavior="hit",
    )


def _scenario_unknown_miss(*, number: int, repeat: int) -> ScenarioResult:
    class_keys = _class_keys(64)
    alias_key = list[int]
    unknown_dependency = object()

    before_dispatch = _compile_identity_dispatch((*class_keys, alias_key))
    after_dispatch = _compile_hybrid_dispatch(
        identity_keys=class_keys,
        equality_keys=(alias_key,),
    )

    def before_call() -> None:
        try:
            before_dispatch(unknown_dependency)
        except KeyError:
            return
        msg = "Unknown dependency unexpectedly resolved in identity dispatch."
        raise RuntimeError(msg)

    def after_call() -> None:
        try:
            after_dispatch(unknown_dependency)
        except KeyError:
            return
        msg = "Unknown dependency unexpectedly resolved in hybrid dispatch."
        raise RuntimeError(msg)

    before_ns, after_ns = _measure_pair_ns_per_op(
        before_callable=before_call,
        after_callable=after_call,
        number=number,
        repeat=repeat,
    )
    return ScenarioResult(
        name="unknown_miss_mixed_n65",
        before_ns_op=before_ns,
        after_ns_op=after_ns,
        ratio_after_over_before=after_ns / before_ns,
        before_behavior="miss",
        after_behavior="miss",
    )


def _collect_results(*, number: int, repeat: int) -> list[ScenarioResult]:
    results: list[ScenarioResult] = []
    for provider_count in (8, 32, 128):
        indices = {
            "first": 0,
            "middle": provider_count // 2,
            "last": provider_count - 1,
        }
        for position_name, index in indices.items():
            results.append(
                _scenario_class_path(
                    provider_count=provider_count,
                    position_name=position_name,
                    index=index,
                    number=number,
                    repeat=repeat,
                ),
            )

    alias_factories = {
        "list_int": lambda: list[int],
        "dict_str_int": lambda: dict[str, int],
        "tuple_int_var": lambda: tuple[int, ...],
        "int_or_str": lambda: int | str,
    }
    for alias_name, alias_factory in alias_factories.items():
        results.append(
            _scenario_alias_equal_not_identical(
                alias_name=alias_name,
                alias_factory=alias_factory,
                number=number,
                repeat=repeat,
            ),
        )

    results.append(_scenario_unknown_miss(number=number, repeat=repeat))
    return results


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output",
        type=Path,
        required=True,
    )
    parser.add_argument(
        "--number",
        type=int,
        default=200_000,
    )
    parser.add_argument(
        "--repeat",
        type=int,
        default=7,
    )
    args = parser.parse_args()

    results = _collect_results(number=args.number, repeat=args.repeat)
    payload = {
        "generated_at": datetime.now(timezone.utc).isoformat(),
        "python_version": platform.python_version(),
        "python_implementation": platform.python_implementation(),
        "platform": platform.platform(),
        "number": args.number,
        "repeat": args.repeat,
        "results": [asdict(result) for result in results],
        "command": " ".join([*sys.argv]),
    }
    args.output.parent.mkdir(parents=True, exist_ok=True)
    args.output.write_text(json.dumps(payload, indent=2) + "\n")


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Callable

from mypy.nodes import ARG_NAMED, ARG_NAMED_OPT, ARG_OPT, ARG_POS, ARG_STAR2, ArgKind
from mypy.plugin import MethodContext, Plugin
from mypy.subtypes import is_same_type
from mypy.types import AnyType, CallableType, Type, TypeOfAny, UnionType, get_proper_type

_INJECT_METHOD_FULLNAME = "diwire._internal.resolver_context.ResolverContext.inject"
_INJECT_DECORATOR_CALL_FULLNAME = "diwire._internal.resolver_context._InjectDecorator.__call__"
_INJECT_RESOLVER_KWARG = "diwire_resolver"
_RESOLVER_PROTOCOL_FULLNAME = "diwire.ResolverProtocol"
_INJECTED_UNION_ITEM_COUNT = 2


class DIWireMypyPlugin(Plugin):
    """Mypy plugin that restores precise typing for ``resolver_context.inject``."""

    def get_method_hook(
        self,
        fullname: str,
    ) -> Callable[[MethodContext], Type] | None:
        if fullname == _INJECT_METHOD_FULLNAME:
            return _transform_inject_method
        if fullname == _INJECT_DECORATOR_CALL_FULLNAME:
            return _transform_inject_decorator_call
        return None


def _transform_inject_method(ctx: MethodContext) -> Type:
    callable_type = _first_argument_callable_type(ctx)
    if callable_type is None:
        return ctx.default_return_type
    return _build_precise_injected_callable_type(ctx=ctx, callable_type=callable_type)


def _transform_inject_decorator_call(ctx: MethodContext) -> Type:
    callable_type = _first_argument_callable_type(ctx)
    if callable_type is None:
        return ctx.default_return_type
    return _build_precise_injected_callable_type(ctx=ctx, callable_type=callable_type)


def _first_argument_callable_type(ctx: MethodContext) -> CallableType | None:
    if not ctx.arg_types or not ctx.arg_types[0]:
        return None
    first_argument_type = get_proper_type(ctx.arg_types[0][0])
    if not isinstance(first_argument_type, CallableType):
        return None
    return first_argument_type


def _build_precise_injected_callable_type(
    *,
    ctx: MethodContext,
    callable_type: CallableType,
) -> CallableType:
    arg_types: list[Type] = []
    arg_kinds: list[ArgKind] = []
    arg_names: list[str | None] = []

    for arg_type, arg_kind, arg_name in zip(
        callable_type.arg_types,
        callable_type.arg_kinds,
        callable_type.arg_names,
        strict=True,
    ):
        parameter_type, is_injected = _unwrap_injected_parameter_type(arg_type)
        arg_types.append(parameter_type)
        arg_kinds.append(_to_optional_argument_kind(arg_kind) if is_injected else arg_kind)
        arg_names.append(arg_name)

    if _INJECT_RESOLVER_KWARG not in arg_names:
        resolver_type = _resolver_protocol_type(ctx)
        insert_at = len(arg_kinds)
        for index, arg_kind in enumerate(arg_kinds):
            if arg_kind == ARG_STAR2:
                insert_at = index
                break

        arg_types.insert(insert_at, resolver_type)
        arg_kinds.insert(insert_at, ARG_NAMED_OPT)
        arg_names.insert(insert_at, _INJECT_RESOLVER_KWARG)

    return callable_type.copy_modified(
        arg_types=arg_types,
        arg_kinds=arg_kinds,
        arg_names=arg_names,
    )


def _unwrap_injected_parameter_type(arg_type: Type) -> tuple[Type, bool]:
    proper_type = get_proper_type(arg_type)
    if not isinstance(proper_type, UnionType):
        return arg_type, False
    if len(proper_type.items) != _INJECTED_UNION_ITEM_COUNT:
        return arg_type, False

    left_item, right_item = proper_type.items
    if not is_same_type(left_item, right_item):
        return arg_type, False

    return left_item, True


def _to_optional_argument_kind(arg_kind: ArgKind) -> ArgKind:
    if arg_kind == ARG_POS:
        return ARG_OPT
    if arg_kind == ARG_NAMED:
        return ARG_NAMED_OPT
    return arg_kind


def _resolver_protocol_type(ctx: MethodContext) -> Type:
    try:
        return ctx.api.named_generic_type(_RESOLVER_PROTOCOL_FULLNAME, [])
    except (AssertionError, KeyError):
        return AnyType(TypeOfAny.from_error)



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from typing import Annotated, Generic, TypeVar

import pytest

from diwire import Component, Container, Lifetime, Scope
from diwire.exceptions import DIWireInvalidRegistrationError


class _Service:
    def __init__(self, value: str) -> None:
        self.value = value


class _ServiceImpl(_Service):
    def __init__(self) -> None:
        super().__init__("impl")


class _Repo:
    pass


class _RepoImpl(_Repo):
    pass


class _RepoDecorator(_Repo):
    def __init__(self, inner: _Repo) -> None:
        self.inner = inner


T = TypeVar("T")


class _GenericValue(Generic[T]):
    def __init__(self, value_type: type[T]) -> None:
        self.value_type = value_type


def test_add_instance_component_registers_component_qualified_key() -> None:
    container = Container()
    expected = _Service("instance")
    marker_expected = _Service("marker")

    container.add_instance(expected, provides=_Service, component="primary")
    container.add_instance(marker_expected, provides=_Service, component=Component("marker"))

    assert container.resolve(Annotated[_Service, Component("primary")]) is expected
    assert container.resolve(Annotated[_Service, Component("marker")]) is marker_expected


def test_add_component_registers_component_qualified_key() -> None:
    container = Container()
    container.add(_ServiceImpl, provides=_Service, component="primary")

    resolved = container.resolve(Annotated[_Service, Component("primary")])
    assert isinstance(resolved, _ServiceImpl)


def test_add_factory_infer_form_supports_component() -> None:
    container = Container()

    def _build_direct() -> _Service:
        return _Service("direct")

    def _build_component() -> _Service:
        return _Service("decorator")

    container.add_factory(_build_direct, component="direct")
    container.add_factory(_build_component, component="decorator")

    resolved_direct = container.resolve(Annotated[_Service, Component("direct")])
    resolved_decorator = container.resolve(Annotated[_Service, Component("decorator")])

    assert resolved_direct.value == "direct"
    assert resolved_decorator.value == "decorator"


def test_add_generator_component_registers_and_runs_cleanup() -> None:
    container = Container()
    cleanup_events: list[str] = []

    def _provide() -> Generator[_Service, None, None]:
        try:
            yield _Service("generator")
        finally:
            cleanup_events.append("closed")

    container.add_generator(
        _provide,
        provides=_Service,
        component="request",
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST) as resolver:
        resolved = resolver.resolve(Annotated[_Service, Component("request")])
        assert resolved.value == "generator"

    assert cleanup_events == ["closed"]


def test_add_context_manager_component_registers_and_runs_cleanup() -> None:
    container = Container()
    cleanup_events: list[str] = []

    @contextmanager
    def _provide() -> Generator[_Service, None, None]:
        try:
            yield _Service("context-manager")
        finally:
            cleanup_events.append("closed")

    container.add_context_manager(
        _provide,
        provides=_Service,
        component="request",
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST) as resolver:
        resolved = resolver.resolve(Annotated[_Service, Component("request")])
        assert resolved.value == "context-manager"

    assert cleanup_events == ["closed"]


def test_decorate_with_component_targets_component_qualified_binding() -> None:
    container = Container()
    container.add(_RepoImpl, provides=_Repo)
    container.add(_RepoImpl, provides=_Repo, component="primary")

    container.decorate(
        provides=_Repo,
        component="primary",
        decorator=_RepoDecorator,
        inner_parameter="inner",
    )

    base = container.resolve(_Repo)
    component_key = Annotated[_Repo, Component("primary")]
    component_resolved = container.resolve(component_key)

    assert isinstance(base, _RepoImpl)
    assert isinstance(component_resolved, _RepoDecorator)
    assert isinstance(component_resolved.inner, _RepoImpl)


def test_component_registration_strips_non_component_metadata_from_provides() -> None:
    container = Container()
    container.add_instance(
        _Service("annotated"),
        provides=Annotated[_Service, "metadata"],
        component="primary",
    )

    resolved = container.resolve(Annotated[_Service, Component("primary")])
    assert resolved.value == "annotated"


def test_non_component_annotated_registration_collides_with_base_key_last_write_wins() -> None:
    container = Container()
    container.add_instance(_Service("base"), provides=_Service)
    container.add_instance(_Service("annotated"), provides=Annotated[_Service, "metadata"])

    resolved = container.resolve(_Service)
    assert resolved.value == "annotated"


def test_non_component_annotated_key_resolves_via_normalized_lookup() -> None:
    container = Container()
    container.add_instance(_Service("base"), provides=_Service)

    resolved = container.resolve(Annotated[_Service, "metadata"])
    assert resolved.value == "base"


@pytest.mark.asyncio
async def test_non_component_annotated_key_aresolves_via_normalized_lookup() -> None:
    container = Container()
    container.add_instance(_Service("base"), provides=_Service)

    resolved = await container.aresolve(Annotated[_Service, "metadata"])
    assert resolved.value == "base"


def test_component_registration_rejects_already_component_qualified_provides() -> None:
    container = Container()
    component_key = Annotated[_Service, Component("existing")]

    with pytest.raises(
        DIWireInvalidRegistrationError,
        match="component-qualified 'provides' key",
    ):
        container.add_factory(lambda: _Service("value"), provides=component_key, component="new")


def test_decorate_rejects_component_when_provides_is_already_component_qualified() -> None:
    container = Container()
    component_key = Annotated[_Repo, Component("existing")]

    with pytest.raises(
        DIWireInvalidRegistrationError,
        match="component-qualified 'provides' key",
    ):
        container.decorate(
            provides=component_key,
            component="new",
            decorator=_RepoDecorator,
            inner_parameter="inner",
        )


def test_component_registration_requires_hashable_component_value() -> None:
    container = Container()

    with pytest.raises(
        DIWireInvalidRegistrationError,
        match="parameter 'component' must be hashable",
    ):
        container.add_instance(_Service("value"), provides=_Service, component=[])

    with pytest.raises(
        DIWireInvalidRegistrationError,
        match="parameter 'component' must be hashable",
    ):
        container.add_instance(_Service("value"), provides=_Service, component=Component([]))


def test_closed_generic_typevar_injection_works_with_component_qualified_provides() -> None:
    container = Container()
    container.add(
        _GenericValue,
        provides=_GenericValue[int],
        component="primary",
    )

    resolved = container.resolve(Annotated[_GenericValue[int], Component("primary")])
    assert resolved.value_type is int



from __future__ import annotations

import inspect
import threading
from collections.abc import AsyncGenerator, Awaitable, Callable, Coroutine, Generator
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from dataclasses import dataclass, field
from enum import Enum, auto
from inspect import Parameter
from types import TracebackType
from typing import (
    Annotated,
    Any,
    ClassVar,
    Literal,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

from diwire._internal.lock_mode import LockMode
from diwire._internal.scope import BaseScope
from diwire.exceptions import (
    DIWireInvalidProviderSpecError,
    DIWireInvalidRegistrationError,
    DIWireProviderDependencyInferenceError,
)

T = TypeVar("T")
_CMT_co = TypeVar("_CMT_co", covariant=True)

UserDependency: TypeAlias = Any
"""A dependency that been registered or trying to be resolved from the user's code."""

UserProviderObject: TypeAlias = Any
"""An object, function, or class provided by the user as a provider. Determined by ProviderKind."""

ProviderSlot: TypeAlias = int
"""A unique slot number assigned to each provider specification."""

ConcreteTypeProvider: TypeAlias = type[T]
"""A concrete type that can be instantiated to produce a dependency."""

FactoryProvider: TypeAlias = Callable[..., T] | Callable[..., Awaitable[T]]
"""A factory function or asynchronous function that produces a dependency."""

GeneratorProvider: TypeAlias = (
    Callable[..., Generator[T, None, None]] | Callable[..., AsyncGenerator[T, None]]
)
"""A generator function or asynchronous generator function that yields a dependency."""


class _ContextManagerLike(Protocol[_CMT_co]):
    def __enter__(self) -> _CMT_co: ...

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None: ...


class _AsyncContextManagerLike(Protocol[_CMT_co]):
    def __aenter__(self) -> Awaitable[_CMT_co]: ...

    def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> Awaitable[bool | None]: ...


ContextManagerProvider: TypeAlias = (
    Callable[..., _ContextManagerLike[T]] | Callable[..., _AsyncContextManagerLike[T]]
)

_MISSING_ANNOTATION: Any = object()
_IMPLICIT_FIRST_PARAMETER_NAMES = {"self", "cls"}
_COROUTINE_RESULT_INDEX = 2
_COROUTINE_ARGUMENT_COUNT = 3
_ASYNC_RESOLVER_ORIGINS: tuple[type[Any], ...] = (
    Awaitable,
    Coroutine,
    AsyncGenerator,
    AbstractAsyncContextManager,
)


@dataclass(kw_only=True)
class ProviderSpec:
    """Describe how a single dependency key is produced and cached.

    A spec is the normalized registration model used by resolver generation.
    Exactly one provider source is expected in user-facing registrations
    (instance, concrete type, factory, generator, or context manager), plus
    metadata describing scope, lifetime, async requirements, cleanup behavior,
    and locking.
    """

    SLOT_COUNTER: ClassVar[ProviderSlot] = 0
    SLOT_COUNTER_LOCK: ClassVar[Any] = threading.Lock()

    provides: UserDependency
    """The dependency type that this provider supplies."""

    instance: UserDependency | None = None
    """An optional instance of the provided dependency, if applicable."""
    concrete_type: ConcreteTypeProvider[Any] | None = None
    """An optional concrete type of the provided dependency, if applicable."""
    factory: FactoryProvider[Any] | None = None
    """An optional factory function to create the provided dependency, if applicable."""
    generator: GeneratorProvider[Any] | None = None
    """An optional generator function to yield the provided dependency, if applicable."""
    context_manager: ContextManagerProvider[Any] | None = None
    """An optional context manager function to manage the provided dependency, if applicable."""
    dependencies: list[ProviderDependency] = field(default_factory=list)
    """A list of dependencies required by this provider."""
    is_async: bool
    """Indicates whether the provider specification itself is asynchronous."""
    is_any_dependency_async: bool
    """Indicates whether any dependency requires asynchronous resolution."""
    needs_cleanup: bool
    """True if provider itself or any dependency requires cleanup."""
    lock_mode: LockMode | Literal["auto"] = "auto"
    """Resolved locking strategy for this provider."""

    lifetime: Lifetime | None = None
    """The lifetime of the provided dependency. Could be none in case of instance providers."""
    scope: BaseScope
    """The scope in which the provided dependency is valid."""

    slot: ProviderSlot = field(init=False)
    """A unique slot number assigned to this provider specification."""

    def __post_init__(self) -> None:
        with self.SLOT_COUNTER_LOCK:
            self.__class__.SLOT_COUNTER += 1
            self.slot = self.SLOT_COUNTER


class Lifetime(Enum):
    """Define cache behavior for provider results."""

    TRANSIENT = auto()
    """Disable caching and build a new value for every resolution call."""

    SCOPED = auto()
    """Cache per declared scope owner until that scope exits.

    For registrations declared on the container root scope, this behaves like a
    singleton for the container lifetime. Cleanup-enabled providers are released
    when the owning scope/resolver is closed.
    """


class ProvidersRegistrations:
    """Store provider specs indexed by dependency key and slot.

    Registration keys are unique: adding a spec for an existing dependency key
    replaces the previous spec. Slot indexing is maintained for resolver
    planning, and cleanup flags are recomputed after each mutation.
    """

    def __init__(self) -> None:
        self._registrations_by_type: dict[UserDependency, ProviderSpec] = {}
        self._registrations_by_slot: dict[int, ProviderSpec] = {}

    @dataclass(frozen=True, slots=True)
    class Snapshot:
        """Capture provider registration state for transactional rollback."""

        registrations_by_type: dict[UserDependency, ProviderSpec]
        registrations_by_slot: dict[int, ProviderSpec]

    def snapshot(self) -> Snapshot:
        """Capture current registrations for rollback."""
        return self.Snapshot(
            registrations_by_type=dict(self._registrations_by_type),
            registrations_by_slot=dict(self._registrations_by_slot),
        )

    def restore(self, snapshot: Snapshot) -> None:
        """Restore registrations from a previous snapshot.

        Args:
            snapshot: Previously captured snapshot state to restore into the registry.

        """
        self._registrations_by_type = dict(snapshot.registrations_by_type)
        self._registrations_by_slot = dict(snapshot.registrations_by_slot)
        self._refresh_needs_cleanup_flags()

    def add(self, spec: ProviderSpec) -> None:
        """Add a new provider specification to the registrations.

        Args:
            spec: Provider specification to register.

        """
        if previous_spec := self._registrations_by_type.get(spec.provides):
            self._registrations_by_slot.pop(previous_spec.slot, None)
        self._registrations_by_type[spec.provides] = spec
        self._registrations_by_slot[spec.slot] = spec
        self._refresh_needs_cleanup_flags()

    def get_by_type(self, dep_type: UserDependency) -> ProviderSpec:
        """Get a provider specification by the type of dependency it provides.

        Args:
            dep_type: Dependency type key to look up.

        """
        return self._registrations_by_type[dep_type]

    def find_by_type(self, dep_type: UserDependency) -> ProviderSpec | None:
        """Get a provider specification by dependency type, if it exists.

        Args:
            dep_type: Dependency type key to look up.

        """
        return self._registrations_by_type.get(dep_type)

    def get_by_slot(self, slot: int) -> ProviderSpec:
        """Get a provider specification by its unique slot number.

        Args:
            slot: Provider slot identifier to look up.

        """
        return self._registrations_by_slot[slot]

    def get_by_scope(self, scope: BaseScope | None) -> list[ProviderSpec]:
        """Get all provider specifications registered for a specific scope.

        Args:
            scope: Scope value used to filter registrations or open nested resolution scope.

        """
        return [spec for spec in self._registrations_by_type.values() if spec.scope == scope]

    def values(self) -> list[ProviderSpec]:
        """Get all provider specifications."""
        return list(self._registrations_by_type.values())

    def _refresh_needs_cleanup_flags(self) -> None:
        """Recompute cleanup requirements for all registered providers."""
        # Dependencies can be registered after dependents, so recompute until stable.
        has_changes = True
        while has_changes:
            has_changes = False
            for spec in self.values():
                dependency_needs_cleanup = any(
                    dependency_spec.needs_cleanup
                    for dependency in spec.dependencies
                    if (dependency_spec := self.find_by_type(dependency.provides))
                )
                needs_cleanup = (
                    spec.generator is not None
                    or spec.context_manager is not None
                    or dependency_needs_cleanup
                )
                if spec.needs_cleanup is not needs_cleanup:
                    spec.needs_cleanup = needs_cleanup
                    has_changes = True

    def __len__(self) -> int:
        return len(self._registrations_by_type)


@dataclass(slots=True)
class ProviderDependenciesExtractor:
    """Extracts dependencies from user-defined provider objects."""

    def extract_from_concrete_type(
        self,
        concrete_type: ConcreteTypeProvider[Any],
    ) -> list[ProviderDependency]:
        """Extract dependencies from a concrete type-based provider.

        Args:
            concrete_type: Concrete class provider to inspect or validate.

        """
        return self._extract_dependencies(
            provider=concrete_type,
            provider_name=concrete_type.__qualname__,
            skip_first_parameter=False,
        )

    def extract_from_factory(
        self,
        factory: FactoryProvider[Any],
    ) -> list[ProviderDependency]:
        """Extract dependencies from a factory-based provider.

        Args:
            factory: Factory provider callable to inspect or validate.

        """
        return self._extract_dependencies(
            provider=factory,
            provider_name=self._provider_name(factory),
            skip_first_parameter=False,
        )

    def extract_from_generator(
        self,
        generator: GeneratorProvider[Any],
    ) -> list[ProviderDependency]:
        """Extract dependencies from a generator-based provider.

        Args:
            generator: Generator provider callable to inspect or validate.

        """
        return self._extract_dependencies(
            provider=generator,
            provider_name=self._provider_name(generator),
            skip_first_parameter=False,
        )

    def extract_from_context_manager(
        self,
        context_manager: ContextManagerProvider[Any],
    ) -> list[ProviderDependency]:
        """Extract dependencies from a context manager-based provider.

        Args:
            context_manager: Context manager provider callable to inspect or validate.

        """
        return self._extract_dependencies(
            provider=context_manager,
            provider_name=self._provider_name(context_manager),
            skip_first_parameter=False,
        )

    def validate_explicit_for_concrete_type(
        self,
        concrete_type: ConcreteTypeProvider[Any],
        dependencies: list[ProviderDependency],
    ) -> list[ProviderDependency]:
        """Validate explicit dependencies for a concrete type-based provider.

        Args:
            concrete_type: Concrete class provider to inspect or validate.
            dependencies: Explicit dependency declarations to validate or inspect for async requirements.

        """
        return self._validate_explicit_dependencies(
            provider=concrete_type,
            provider_name=concrete_type.__qualname__,
            dependencies=dependencies,
            skip_first_parameter=False,
        )

    def validate_explicit_for_factory(
        self,
        factory: FactoryProvider[Any],
        dependencies: list[ProviderDependency],
    ) -> list[ProviderDependency]:
        """Validate explicit dependencies for a factory-based provider.

        Args:
            factory: Factory provider callable to inspect or validate.
            dependencies: Explicit dependency declarations to validate or inspect for async requirements.

        """
        return self._validate_explicit_dependencies(
            provider=factory,
            provider_name=self._provider_name(factory),
            dependencies=dependencies,
            skip_first_parameter=False,
        )

    def validate_explicit_for_generator(
        self,
        generator: GeneratorProvider[Any],
        dependencies: list[ProviderDependency],
    ) -> list[ProviderDependency]:
        """Validate explicit dependencies for a generator-based provider.

        Args:
            generator: Generator provider callable to inspect or validate.
            dependencies: Explicit dependency declarations to validate or inspect for async requirements.

        """
        return self._validate_explicit_dependencies(
            provider=generator,
            provider_name=self._provider_name(generator),
            dependencies=dependencies,
            skip_first_parameter=False,
        )

    def validate_explicit_for_context_manager(
        self,
        context_manager: ContextManagerProvider[Any],
        dependencies: list[ProviderDependency],
    ) -> list[ProviderDependency]:
        """Validate explicit dependencies for a context manager-based provider.

        Args:
            context_manager: Context manager provider callable to inspect or validate.
            dependencies: Explicit dependency declarations to validate or inspect for async requirements.

        """
        return self._validate_explicit_dependencies(
            provider=context_manager,
            provider_name=self._provider_name(context_manager),
            dependencies=dependencies,
            skip_first_parameter=False,
        )

    def _extract_dependencies(
        self,
        *,
        provider: Callable[..., Any],
        provider_name: str,
        skip_first_parameter: bool,
    ) -> list[ProviderDependency]:
        parameters = self._provider_parameters(
            provider=provider,
            skip_first_parameter=skip_first_parameter,
        )
        annotations, annotation_error = self._resolved_type_hints(provider)
        dependencies: list[ProviderDependency] = []

        for parameter in parameters:
            provides = self._resolve_parameter_annotation(
                parameter=parameter,
                annotations=annotations,
                annotation_error=annotation_error,
                provider_name=provider_name,
            )
            if provides is _MISSING_ANNOTATION:
                continue

            dependencies.append(
                ProviderDependency(
                    provides=provides,
                    parameter=parameter,
                ),
            )

        return dependencies

    def _validate_explicit_dependencies(
        self,
        *,
        provider: Callable[..., Any],
        provider_name: str,
        dependencies: list[ProviderDependency],
        skip_first_parameter: bool,
    ) -> list[ProviderDependency]:
        parameters = self._provider_parameters(
            provider=provider,
            skip_first_parameter=skip_first_parameter,
        )
        parameters_by_name = {parameter.name: parameter for parameter in parameters}
        validated_dependencies_by_name: dict[str, ProviderDependency] = {}

        for dependency in dependencies:
            parameter_name = dependency.parameter.name
            signature_parameter = parameters_by_name.get(parameter_name)
            if signature_parameter is None:
                msg = (
                    f"Explicit dependency for unknown parameter '{parameter_name}' in "
                    f"provider '{provider_name}'."
                )
                raise DIWireInvalidProviderSpecError(msg)
            if parameter_name in validated_dependencies_by_name:
                msg = (
                    f"Explicit dependency for parameter '{parameter_name}' in "
                    f"provider '{provider_name}' is duplicated."
                )
                raise DIWireInvalidProviderSpecError(msg)
            if dependency.parameter.kind is not signature_parameter.kind:
                msg = (
                    f"Explicit dependency for parameter '{parameter_name}' in provider "
                    f"'{provider_name}' has kind '{dependency.parameter.kind}' but expected "
                    f"'{signature_parameter.kind}'."
                )
                raise DIWireInvalidProviderSpecError(msg)

            validated_dependencies_by_name[parameter_name] = ProviderDependency(
                provides=dependency.provides,
                parameter=signature_parameter,
            )

        missing_required_parameters = [
            parameter.name
            for parameter in parameters
            if self._is_required_parameter(parameter)
            and parameter.name not in validated_dependencies_by_name
        ]
        if missing_required_parameters:
            missing_parameters = ", ".join(f"'{name}'" for name in missing_required_parameters)
            msg = (
                f"Explicit dependencies for provider '{provider_name}' are incomplete. "
                f"Missing required parameters: {missing_parameters}."
            )
            raise DIWireProviderDependencyInferenceError(msg)

        return [
            validated_dependencies_by_name[parameter.name]
            for parameter in parameters
            if parameter.name in validated_dependencies_by_name
        ]

    def _resolve_parameter_annotation(
        self,
        *,
        parameter: Parameter,
        annotations: dict[str, Any],
        annotation_error: Exception | None,
        provider_name: str,
    ) -> Any:
        annotation = annotations.get(parameter.name, _MISSING_ANNOTATION)
        if annotation is not _MISSING_ANNOTATION:
            return annotation

        raw_annotation = parameter.annotation
        if raw_annotation is not Parameter.empty and not isinstance(raw_annotation, str):
            return raw_annotation

        if not self._is_required_parameter(parameter):
            return _MISSING_ANNOTATION

        error_message = (
            f"Unable to infer dependency for required parameter '{parameter.name}' "
            f"in provider '{provider_name}'. Add a type annotation or pass explicit dependencies."
        )
        if annotation_error is None:
            raise DIWireProviderDependencyInferenceError(error_message)
        msg = f"{error_message} Original annotation error: {annotation_error}"
        raise DIWireProviderDependencyInferenceError(msg) from annotation_error

    def _provider_parameters(
        self,
        *,
        provider: Callable[..., Any],
        skip_first_parameter: bool,
    ) -> tuple[Parameter, ...]:
        parameters = tuple(inspect.signature(provider).parameters.values())
        if (
            skip_first_parameter
            and parameters
            and parameters[0].name in _IMPLICIT_FIRST_PARAMETER_NAMES
        ):
            return parameters[1:]
        return parameters

    def _resolved_type_hints(
        self,
        provider: Callable[..., Any],
    ) -> tuple[dict[str, Any], Exception | None]:
        annotations: dict[str, Any] = {}
        annotation_error: Exception | None = None

        try:
            annotations = get_type_hints(provider, include_extras=True)
        except (AttributeError, NameError, TypeError) as error:
            annotation_error = error

        if inspect.isclass(provider):
            annotations, annotation_error = self._merge_concrete_type_callable_hints(
                concrete_type=provider,
                annotations=annotations,
                annotation_error=annotation_error,
            )

        return annotations, annotation_error

    def _merge_concrete_type_callable_hints(
        self,
        *,
        concrete_type: type[Any],
        annotations: dict[str, Any],
        annotation_error: Exception | None,
    ) -> tuple[dict[str, Any], Exception | None]:
        merged_annotations = dict(annotations)
        merged_error = annotation_error

        for callable_member_name in ("__call__", "__new__", "__init__"):
            callable_member = getattr(concrete_type, callable_member_name)
            try:
                member_annotations = get_type_hints(callable_member, include_extras=True)
            except (AttributeError, NameError, TypeError) as error:
                if merged_error is None:
                    merged_error = error
                continue
            for parameter_name, parameter_annotation in member_annotations.items():
                merged_annotations.setdefault(parameter_name, parameter_annotation)

        return merged_annotations, merged_error

    def _is_required_parameter(self, parameter: Parameter) -> bool:
        return (
            parameter.default is Parameter.empty
            and parameter.kind is not Parameter.VAR_POSITIONAL
            and parameter.kind is not Parameter.VAR_KEYWORD
        )

    def _provider_name(self, provider: Callable[..., Any]) -> str:
        return getattr(provider, "__qualname__", repr(provider))


@dataclass(slots=True)
class ProviderDependency:
    """Represent a resolved dependency key bound to a provider parameter."""

    provides: UserDependency
    parameter: Parameter


@dataclass(slots=True)
class ProviderReturnTypeExtractor:
    """Extracts return types from user-defined provider objects."""

    def is_factory_async(
        self,
        factory: FactoryProvider[Any],
    ) -> bool:
        """Check whether a factory provider itself is asynchronous.

        Args:
            factory: Factory provider callable to inspect or validate.

        """
        return inspect.iscoroutinefunction(factory) or self.return_annotation_matches_origins(
            provider=factory,
            expected_origins=(Awaitable, Coroutine),
        )

    def is_generator_async(
        self,
        generator: GeneratorProvider[Any],
    ) -> bool:
        """Check whether a generator provider itself is asynchronous.

        Args:
            generator: Generator provider callable to inspect or validate.

        """
        return inspect.isasyncgenfunction(generator) or self.return_annotation_matches_origins(
            provider=generator,
            expected_origins=(AsyncGenerator,),
        )

    def is_context_manager_async(
        self,
        context_manager: ContextManagerProvider[Any],
    ) -> bool:
        """Check whether a context manager provider itself is asynchronous.

        Args:
            context_manager: Context manager provider callable to inspect or validate.

        """
        unwrapped_context_manager = inspect.unwrap(context_manager)
        if inspect.isasyncgenfunction(unwrapped_context_manager):
            return True
        if inspect.iscoroutinefunction(unwrapped_context_manager):
            return True
        if self.return_annotation_matches_origins(
            provider=unwrapped_context_manager,
            expected_origins=(AbstractAsyncContextManager, AsyncGenerator),
        ):
            return True

        if isinstance(unwrapped_context_manager, type) and (
            self._context_manager_mode(unwrapped_context_manager) == "async"
        ):
            return True

        return_annotation, _annotation_error = self._resolved_return_annotation(
            unwrapped_context_manager,
        )
        unwrapped_return_annotation = self.unwrap_annotated(return_annotation)
        if isinstance(unwrapped_return_annotation, type):
            return self._context_manager_mode(unwrapped_return_annotation) == "async"
        return False

    def is_any_dependency_async(
        self,
        dependencies: list[ProviderDependency],
    ) -> bool:
        """Check whether provider dependencies require asynchronous resolution.

        Args:
            dependencies: Explicit dependency declarations to validate or inspect for async requirements.

        """
        return any(
            self.annotation_matches_origins(
                annotation=dependency.provides,
                expected_origins=_ASYNC_RESOLVER_ORIGINS,
            )
            for dependency in dependencies
        )

    def extract_from_factory(
        self,
        factory: FactoryProvider[Any],
    ) -> Any:
        """Extract a return type from a factory-based provider.

        Args:
            factory: Factory provider callable to inspect or validate.

        """
        return_annotation, annotation_error = self._resolved_return_annotation(factory)
        provider_name = self._provider_name(factory)
        if return_annotation is _MISSING_ANNOTATION:
            self._raise_missing_return_annotation_error(
                provider_kind="factory",
                provider_name=provider_name,
                annotation_error=annotation_error,
            )

        unwrapped_return_type = self._unwrap_factory_return_type(return_annotation)
        if unwrapped_return_type is _MISSING_ANNOTATION:
            self._raise_invalid_return_annotation_error(
                provider_kind="factory",
                provider_name=provider_name,
                expected_annotations="`T`, `Awaitable[T]`, or `Coroutine[Any, Any, T]`",
                annotation_error=annotation_error,
            )

        return unwrapped_return_type

    def extract_from_generator(
        self,
        generator: GeneratorProvider[Any],
    ) -> Any:
        """Extract a return type from a generator-based provider.

        Args:
            generator: Generator provider callable to inspect or validate.

        """
        return_annotation, annotation_error = self._resolved_return_annotation(generator)
        provider_name = self._provider_name(generator)
        if return_annotation is _MISSING_ANNOTATION:
            self._raise_missing_return_annotation_error(
                provider_kind="generator",
                provider_name=provider_name,
                annotation_error=annotation_error,
            )

        yielded_type = self._extract_yielded_or_managed_type(
            return_annotation=return_annotation,
            expected_origins=(Generator, AsyncGenerator),
        )
        if yielded_type is _MISSING_ANNOTATION:
            self._raise_invalid_return_annotation_error(
                provider_kind="generator",
                provider_name=provider_name,
                expected_annotations="`Generator[T, None, None]` or `AsyncGenerator[T, None]`",
                annotation_error=annotation_error,
            )

        return yielded_type

    def extract_from_context_manager(
        self,
        context_manager: ContextManagerProvider[Any],
    ) -> Any:
        """Extract a return type from a context manager-based provider.

        Args:
            context_manager: Context manager provider callable to inspect or validate.

        """
        return_annotation, annotation_error = self._resolved_return_annotation(context_manager)
        provider_name = self._provider_name(context_manager)
        if return_annotation is not _MISSING_ANNOTATION:
            yielded_type = self._extract_yielded_or_managed_type(
                return_annotation=return_annotation,
                expected_origins=(
                    AbstractContextManager,
                    AbstractAsyncContextManager,
                    Generator,
                    AsyncGenerator,
                ),
            )
            if yielded_type is not _MISSING_ANNOTATION:
                return yielded_type

        if isinstance(context_manager, type):
            managed_type = self._infer_managed_type_from_cm_type(context_manager)
            if managed_type is not _MISSING_ANNOTATION:
                return managed_type

        unwrapped_return_annotation = self.unwrap_annotated(return_annotation)
        if isinstance(unwrapped_return_annotation, type):
            managed_type = self._infer_managed_type_from_cm_type(unwrapped_return_annotation)
            if managed_type is not _MISSING_ANNOTATION:
                return managed_type

        msg = (
            f"Unable to infer return type for context manager provider '{provider_name}'. "
            "Add a provider return annotation (`AbstractContextManager[T]`, "
            "`AbstractAsyncContextManager[T]`, `Generator[T, None, None]`, or "
            "`AsyncGenerator[T, None]`), annotate `__enter__` or `__aenter__` on the "
            "context manager class, or pass provides= explicitly."
        )
        self._raise_invalid_registration_error(msg=msg, annotation_error=annotation_error)
        return _MISSING_ANNOTATION  # pragma: no cover

    def return_annotation_matches_origins(
        self,
        *,
        provider: Callable[..., Any],
        expected_origins: tuple[type[Any], ...],
    ) -> bool:
        """Check whether provider return annotation origin matches one of expected origins.

        Args:
            provider: Provider callable or type whose return annotation is being checked.
            expected_origins: Allowed annotation origins for a successful match check.

        """
        return_annotation, _annotation_error = self._resolved_return_annotation(provider)
        if return_annotation is _MISSING_ANNOTATION:
            return False
        return self.annotation_matches_origins(
            annotation=return_annotation,
            expected_origins=expected_origins,
        )

    def annotation_matches_origins(
        self,
        *,
        annotation: Any,
        expected_origins: tuple[type[Any], ...],
    ) -> bool:
        """Check whether annotation origin matches one of expected origins.

        Args:
            annotation: Annotation value to inspect or normalize.
            expected_origins: Allowed annotation origins for a successful match check.

        """
        annotation = self.unwrap_annotated(annotation)
        origin = get_origin(annotation)
        if origin in expected_origins:
            return True
        return annotation in expected_origins

    def unwrap_annotated(
        self,
        annotation: Any,
    ) -> Any:
        """Recursively unwrap Annotated[T, ...] into T.

        Args:
            annotation: Annotation value to inspect or normalize.

        """
        if get_origin(annotation) is not Annotated:
            return annotation
        annotation_args = get_args(annotation)
        if (
            not annotation_args
        ):  # pragma: no cover - typing.Annotated always wraps at least one type
            return annotation
        return self.unwrap_annotated(annotation_args[0])

    def _is_self_annotation(
        self,
        annotation: Any,
    ) -> bool:
        unwrapped_annotation = self.unwrap_annotated(annotation)
        annotation_module = getattr(unwrapped_annotation, "__module__", None)
        if annotation_module not in {"typing", "typing_extensions"}:
            return False

        annotation_name = getattr(
            unwrapped_annotation,
            "__qualname__",
            getattr(unwrapped_annotation, "__name__", None),
        )
        return annotation_name == "Self"

    def _infer_managed_type_from_cm_type(
        self,
        cm_type: type[Any],
    ) -> Any:
        context_manager_mode = self._context_manager_mode(cm_type)
        if context_manager_mode == "none":
            return _MISSING_ANNOTATION

        enter_method_name = "__aenter__" if context_manager_mode == "async" else "__enter__"
        enter_method = cast("Callable[..., Any]", getattr(cm_type, enter_method_name))
        return_annotation, _annotation_error = self._resolved_return_annotation(enter_method)
        if return_annotation is _MISSING_ANNOTATION:
            return _MISSING_ANNOTATION

        managed_type = return_annotation
        if context_manager_mode == "async":
            managed_type = self._unwrap_factory_return_type(managed_type)
            if managed_type is _MISSING_ANNOTATION:
                return _MISSING_ANNOTATION

        if self._is_self_annotation(managed_type):
            return cm_type
        return managed_type

    def _context_manager_mode(
        self,
        cm_type: type[Any],
    ) -> Literal["sync", "async", "none"]:
        has_sync_methods = callable(getattr(cm_type, "__enter__", None)) and callable(
            getattr(cm_type, "__exit__", None),
        )
        has_async_methods = callable(getattr(cm_type, "__aenter__", None)) and callable(
            getattr(cm_type, "__aexit__", None),
        )

        if has_async_methods and not has_sync_methods:
            return "async"
        if has_sync_methods:
            return "sync"
        return "none"

    def _resolved_return_annotation(
        self,
        provider: Callable[..., Any],
    ) -> tuple[Any, Exception | None]:
        try:
            return_type_hints = get_type_hints(provider, include_extras=True)
            annotation_error: Exception | None = None
        except (AttributeError, NameError, TypeError) as error:
            return_type_hints = {}
            annotation_error = error

        resolved_return_annotation = return_type_hints.get("return", _MISSING_ANNOTATION)
        if resolved_return_annotation is not _MISSING_ANNOTATION:
            return resolved_return_annotation, annotation_error

        try:
            raw_return_annotation = inspect.signature(provider).return_annotation
        except (TypeError, ValueError) as error:
            if annotation_error is None:
                annotation_error = error
            return _MISSING_ANNOTATION, annotation_error

        if raw_return_annotation is inspect.Signature.empty or isinstance(
            raw_return_annotation,
            str,
        ):
            return _MISSING_ANNOTATION, annotation_error

        return raw_return_annotation, annotation_error

    def _unwrap_factory_return_type(
        self,
        return_annotation: Any,
    ) -> Any:
        origin = get_origin(return_annotation)
        annotation_args = get_args(return_annotation)
        if origin is Awaitable:
            if len(annotation_args) != 1:
                return _MISSING_ANNOTATION
            return annotation_args[0]
        if origin is Coroutine:
            if len(annotation_args) != _COROUTINE_ARGUMENT_COUNT:
                return _MISSING_ANNOTATION
            return annotation_args[_COROUTINE_RESULT_INDEX]
        return return_annotation

    def _extract_yielded_or_managed_type(
        self,
        *,
        return_annotation: Any,
        expected_origins: tuple[type[Any], ...],
    ) -> Any:
        origin = get_origin(return_annotation)
        if origin not in expected_origins:
            return _MISSING_ANNOTATION

        annotation_args = get_args(return_annotation)
        if len(annotation_args) < 1:
            return _MISSING_ANNOTATION
        return annotation_args[0]

    def _raise_missing_return_annotation_error(
        self,
        *,
        provider_kind: str,
        provider_name: str,
        annotation_error: Exception | None,
    ) -> None:
        msg = (
            f"Unable to infer return type for {provider_kind} provider '{provider_name}'. "
            "Add a return type annotation or pass provides= explicitly."
        )
        self._raise_invalid_registration_error(msg=msg, annotation_error=annotation_error)

    def _raise_invalid_return_annotation_error(
        self,
        *,
        provider_kind: str,
        provider_name: str,
        expected_annotations: str,
        annotation_error: Exception | None,
    ) -> None:
        msg = (
            f"Unable to infer return type for {provider_kind} provider '{provider_name}'. "
            f"Expected return annotation {expected_annotations}. "
            "Add a valid return annotation or pass provides= explicitly."
        )
        self._raise_invalid_registration_error(msg=msg, annotation_error=annotation_error)

    def _raise_invalid_registration_error(
        self,
        *,
        msg: str,
        annotation_error: Exception | None,
    ) -> None:
        if annotation_error is None:
            raise DIWireInvalidRegistrationError(msg)
        full_msg = f"{msg} Original annotation error: {annotation_error}"
        raise DIWireInvalidRegistrationError(full_msg) from annotation_error

    def _provider_name(self, provider: Callable[..., Any]) -> str:
        return getattr(provider, "__qualname__", repr(provider))



from flask import Flask

from diwire import Container, Injected, Scope, resolver_context
from diwire.integrations.flask import add_request_context
from tests.e2e.flask.setup.config import FlaskE2ESettings
from tests.e2e.flask.setup.services import CMService, RequestBasedService, latest_cleanup_path

app = Flask(__name__)


@app.get("/health")
def health() -> str:
    return "OK"


@app.get("/services/request-based")
@resolver_context.inject(scope=Scope.REQUEST)
def request_based_service(service: Injected[RequestBasedService]) -> str:
    return service.work()


@app.get("/services/cm-service")
@resolver_context.inject(scope=Scope.REQUEST)
def cm_service(service: Injected[CMService]) -> str:
    return service.work()


@app.get("/services/cm-service/cleanup")
def cm_service_cleanup() -> str:
    cleanup_path = latest_cleanup_path()
    return f"Cleanup path: {cleanup_path}"


def main() -> None:
    container = Container()
    add_request_context(container)
    container.add_context_manager(CMService, scope=Scope.REQUEST)

    settings = container.resolve(FlaskE2ESettings)
    app.run(host=settings.host, port=settings.port)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from diwire import Container, Injected, Lifetime, Scope, resolver_context
from diwire.exceptions import DIWireIntegrationError

if TYPE_CHECKING:
    import typer
    from typer.testing import CliRunner

typer = pytest.importorskip("typer")


def _add_typer_context(container: Container) -> None:
    from diwire.integrations.typer import add_typer_context

    add_typer_context(container)


class _ContextInfoService:
    def __init__(self, context: typer.Context) -> None:
        self._context = context

    def describe(self, current: typer.Context) -> str:
        info_name = self._context.info_name or "<missing>"
        same = "same" if self._context is current else "different"
        return f"{info_name}:{same}"


@pytest.fixture()
def app() -> typer.Typer:
    container = Container()
    _add_typer_context(container)
    container.add(
        _ContextInfoService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    app = typer.Typer()

    @app.command("context-direct")
    @resolver_context.inject(scope=Scope.REQUEST)
    def context_direct(
        ctx: typer.Context,
        resolved_ctx: Injected[typer.Context],
    ) -> None:
        info_name = resolved_ctx.info_name or "<missing>"
        same = "same" if resolved_ctx is ctx else "different"
        typer.echo(f"{info_name}:{same}")

    @app.command("context-service")
    @resolver_context.inject(scope=Scope.REQUEST)
    def context_service(
        ctx: typer.Context,
        service: Injected[_ContextInfoService],
    ) -> None:
        typer.echo(service.describe(ctx))

    return app


@pytest.fixture()
def runner() -> CliRunner:
    from typer.testing import CliRunner

    return CliRunner()


def test_context_resolve_for_direct_command(runner: CliRunner, app: typer.Typer) -> None:
    result = runner.invoke(app, ["context-direct"], prog_name="diwire")

    assert result.exit_code == 0
    assert result.stdout.strip() == "context-direct:same"


def test_context_resolve_in_service_for_command(runner: CliRunner, app: typer.Typer) -> None:
    result = runner.invoke(app, ["context-service"], prog_name="diwire")

    assert result.exit_code == 0
    assert result.stdout.strip() == "context-service:same"


def test_context_resolution_raises_when_not_invoked_by_typer() -> None:
    container = Container()
    _add_typer_context(container)

    with pytest.raises(DIWireIntegrationError, match="Typer context not available"):
        container.resolve(typer.Context)



"""Named components with ``Component("name")`` and ``Annotated`` keys.

This module demonstrates how to register multiple implementations of the same
base type and resolve/inject them using component-qualified dependency keys.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated

from diwire import Component, Container, Injected, resolver_context


@dataclass(slots=True)
class UserStore:
    backend: str

    def get_user(self, user_id: int) -> str:
        return f"{self.backend}:user:{user_id}"


PrimaryStore = Annotated[UserStore, Component("primary")]
FallbackStore = Annotated[UserStore, Component("fallback")]


def main() -> None:
    container = Container()

    container.add_factory(lambda: UserStore(backend="redis"), provides=PrimaryStore)
    container.add_factory(lambda: UserStore(backend="memory"), provides=FallbackStore)

    @resolver_context.inject
    def load_users(
        primary_store: Injected[PrimaryStore],
        fallback_store: Injected[FallbackStore],
    ) -> tuple[str, str]:
        return primary_store.get_user(1), fallback_store.get_user(1)

    primary_user, fallback_user = load_users()
    print(f"primary={primary_user}")  # => primary=redis:user:1
    print(f"fallback={fallback_user}")  # => fallback=memory:user:1


if __name__ == "__main__":
    main()



"""ContextVar-backed dependencies work with default values without opening scopes."""

from __future__ import annotations

from contextvars import ContextVar

from diwire import Container, Injected, Lifetime, resolver_context

current_value_var: ContextVar[int] = ContextVar("current_value_no_scope", default=5)


def read_current_value() -> int:
    return current_value_var.get()


def main() -> None:
    container = Container()
    container.add_factory(read_current_value, provides=int, lifetime=Lifetime.TRANSIENT)

    @resolver_context.inject(auto_open_scope=False)
    def handler(value: Injected[int]) -> int:
        return value

    default_value = handler()

    token = current_value_var.set(7)
    try:
        overridden_value = handler()
    finally:
        current_value_var.reset(token)

    print(f"default_value={default_value}")  # => default_value=5
    print(f"overridden_value={overridden_value}")  # => overridden_value=7


if __name__ == "__main__":
    main()



from diwire._internal.providers import ProvidersRegistrations
from diwire._internal.resolvers.assembly.compiler import ResolversAssemblyCompiler
from diwire._internal.resolvers.assembly.planner import validate_resolver_assembly_managed_scopes
from diwire._internal.resolvers.protocol import ResolverProtocol
from diwire._internal.scope import BaseScope


class ResolversManager:
    """Manager for dependency resolvers."""

    def __init__(self) -> None:
        self._assembly_compiler = ResolversAssemblyCompiler()

    def build_root_resolver(
        self,
        root_scope: BaseScope,
        registrations: ProvidersRegistrations,
    ) -> ResolverProtocol:
        """Get the root resolver for the given registrations.

        Generates the resolver code dynamically based on the provided registrations and root scope.

        Args:
            root_scope: Root scope used to initialize the resolver.
            registrations: Provider registrations used to build resolver instances or generated code.

        """
        validate_resolver_assembly_managed_scopes(root_scope=root_scope)
        return self._assembly_compiler.build_root_resolver(
            root_scope=root_scope,
            registrations=registrations,
        )



from __future__ import annotations

import pytest

from diwire import Container, Injected, Lifetime

pytest_plugins = ["diwire.integrations.pytest_plugin"]


class Service:
    pass


class ServiceImpl(Service):
    pass


@pytest.fixture()
def diwire_container() -> Container:
    container = Container()
    container.add(
        ServiceImpl,
        provides=Service,
        lifetime=Lifetime.SCOPED,
    )
    return container


def test_plugin_injects_parameters(service: Injected[Service]) -> None:
    if not isinstance(service, ServiceImpl):
        msg = "Injected service is not ServiceImpl"
        raise TypeError(msg)



from __future__ import annotations

import inspect
from dataclasses import dataclass
from types import TracebackType
from typing import Any, cast
from uuid import uuid4

import pytest
from typing_extensions import Self

from diwire import Container, Injected, Lifetime, Scope, resolver_context


@dataclass(frozen=True)
class _Calculator:
    offset: int

    def add(self, left: int, right: int) -> int:
        return left + right + self.offset


def _build_calculator() -> _Calculator:
    return _Calculator(offset=3)


@dataclass(frozen=True)
class _ScopedToken:
    value: str


def _build_scoped_token() -> _ScopedToken:
    return _ScopedToken(value=uuid4().hex)


@dataclass
class _ScopedCleanupResource:
    closed_values: list[str]
    value: str

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.closed_values.append(self.value)


def test_resolver_context_injected_task_resolves_injected_parameters_for_sync_callable() -> None:
    container = Container()
    container.add_factory(_build_calculator, provides=_Calculator)

    @resolver_context.inject(scope=Scope.REQUEST)
    def task(left: int, right: int, calculator: Injected[_Calculator]) -> int:
        return calculator.add(left, right)

    injected_task = cast("Any", task)
    assert injected_task(5, 7) == 15


def test_resolver_context_injected_task_hides_injected_parameters_in_public_signature() -> None:
    container = Container()
    container.add_factory(_build_calculator, provides=_Calculator)

    @resolver_context.inject(scope=Scope.REQUEST)
    def task(left: int, right: int, calculator: Injected[_Calculator]) -> int:
        return calculator.add(left, right)

    parameter_names = tuple(inspect.signature(task).parameters)
    assert parameter_names == ("left", "right")


def test_resolver_context_injected_task_keeps_explicit_parameter_values() -> None:
    container = Container()
    container.add_factory(_build_calculator, provides=_Calculator)

    @resolver_context.inject(scope=Scope.REQUEST)
    def task(left: int, right: int, calculator: Injected[_Calculator]) -> int:
        return calculator.add(left, right)

    injected_task = cast("Any", task)
    explicit_calculator = _Calculator(offset=100)
    assert injected_task(2, 3, calculator=explicit_calculator) == 105


def test_resolver_context_injected_task_preserves_scope_isolation_per_invocation() -> None:
    container = Container()
    closed_values: list[str] = []
    container.add_factory(
        _build_scoped_token,
        provides=_ScopedToken,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    def _provide_cleanup_resource(token: _ScopedToken) -> _ScopedCleanupResource:
        return _ScopedCleanupResource(closed_values=closed_values, value=token.value)

    container.add_context_manager(
        _provide_cleanup_resource,
        provides=_ScopedCleanupResource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    @resolver_context.inject(scope=Scope.REQUEST)
    def task(
        first: Injected[_ScopedToken],
        second: Injected[_ScopedToken],
        cleanup_resource: Injected[_ScopedCleanupResource],
    ) -> tuple[str, str, str]:
        return (first.value, second.value, cleanup_resource.value)

    injected_task = cast("Any", task)
    first = injected_task()
    second = injected_task()

    assert first[0] == first[1] == first[2]
    assert second[0] == second[1] == second[2]
    assert first[0] != second[0]
    assert closed_values == [first[0], second[0]]


@pytest.mark.asyncio
async def test_resolver_context_injected_task_resolves_injected_parameters_for_async_callable() -> (
    None
):
    container = Container()
    container.add_factory(_build_calculator, provides=_Calculator)

    @resolver_context.inject(scope=Scope.REQUEST)
    async def task(left: int, right: int, calculator: Injected[_Calculator]) -> int:
        return calculator.add(left, right)

    injected_task = cast("Any", task)
    assert await injected_task(4, 9) == 16



from __future__ import annotations

from contextvars import ContextVar

import pytest

from diwire import Container, Lifetime, Scope
from diwire._internal.scope import BaseScope, BaseScopes, Scopes
from diwire.exceptions import DIWireScopeMismatchError


class _RequestOnly:
    pass


class _ActionContextValue:
    def __init__(self, value: int) -> None:
        self.value = value


class _AsyncRequestOnly:
    pass


_action_value_var: ContextVar[int] = ContextVar("action_value", default=0)


def _read_action_value() -> int:
    return _action_value_var.get()


def test_container_resolve_uses_context_bound_scope_by_default() -> None:
    container = Container()
    container.add(
        _RequestOnly,
        provides=_RequestOnly,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST):
        assert isinstance(container.resolve(_RequestOnly), _RequestOnly)


def test_plain_class_attributes_do_not_rebind_shared_scope_metadata() -> None:
    class _ScopeHolder:
        app_scope = Scope.APP
        request_scope = Scope.REQUEST

    assert _ScopeHolder.app_scope is Scope.APP
    assert _ScopeHolder.request_scope is Scope.REQUEST
    assert Scope.APP.owner is Scopes
    assert Scope.REQUEST.owner is Scopes
    assert Scope.APP.scope_name == "APP"
    assert Scope.REQUEST.scope_name == "REQUEST"

    container = Container()
    container.add(
        _RequestOnly,
        provides=_RequestOnly,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST):
        assert isinstance(container.resolve(_RequestOnly), _RequestOnly)


def test_scope_metadata_binding_is_idempotent_for_the_same_owner() -> None:
    scope = BaseScope(99)

    scope.__set_name__(Scopes, "CUSTOM")
    scope.__set_name__(Scopes, "CUSTOM")

    assert scope.owner is Scopes
    assert scope.scope_name == "CUSTOM"


def test_scope_metadata_binding_keeps_first_base_scopes_owner() -> None:
    class _OtherScopes(BaseScopes):
        pass

    scope = BaseScope(100)

    scope.__set_name__(Scopes, "CUSTOM")
    scope.__set_name__(_OtherScopes, "OTHER")

    assert scope.owner is Scopes
    assert scope.scope_name == "CUSTOM"


@pytest.mark.asyncio
async def test_container_aresolve_uses_context_bound_scope_by_default() -> None:
    container = Container()

    async def build() -> _AsyncRequestOnly:
        return _AsyncRequestOnly()

    container.add_factory(
        build,
        provides=_AsyncRequestOnly,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST):
        assert isinstance(await container.aresolve(_AsyncRequestOnly), _AsyncRequestOnly)


def test_container_resolve_does_not_use_context_bound_scope_when_disabled() -> None:
    container = Container(use_resolver_context=False)
    container.add(
        _RequestOnly,
        provides=_RequestOnly,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST):
        with pytest.raises(DIWireScopeMismatchError, match="requires opened scope level"):
            _ = container.resolve(_RequestOnly)


def test_container_enter_scope_uses_context_bound_scope_for_nesting() -> None:
    container = Container()
    container.add_factory(
        _read_action_value,
        provides=int,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add(
        _ActionContextValue,
        provides=_ActionContextValue,
        scope=Scope.ACTION,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST):
        token = _action_value_var.set(7)
        try:
            with container.enter_scope(Scope.ACTION):
                assert container.resolve(_ActionContextValue).value == 7
        finally:
            _action_value_var.reset(token)


def test_container_enter_scope_uses_contextvar_default_when_token_is_not_set() -> None:
    container = Container(use_resolver_context=False)
    container.add_factory(
        _read_action_value,
        provides=int,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add(
        _ActionContextValue,
        provides=_ActionContextValue,
        scope=Scope.ACTION,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.ACTION) as action_resolver:
        assert action_resolver.resolve(_ActionContextValue).value == 0



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, TypeVar, cast

import pytest

from diwire import (
    AsyncProvider,
    Container,
    DependencyRegistrationPolicy,
    Injected,
    Lifetime,
    Provider,
    Scope,
    resolver_context,
)
from diwire.exceptions import (
    DIWireAsyncDependencyInSyncContextError,
    DIWireInvalidProviderSpecError,
)


class _CycleB:
    def __init__(self, a: _CycleA) -> None:
        self.a = a


class _CycleA:
    def __init__(self, b_provider: Provider[_CycleB]) -> None:
        self._b_provider = b_provider

    def get_b(self) -> _CycleB:
        return self._b_provider()


class _DirectCycleA:
    def __init__(self, b: _DirectCycleB) -> None:
        self.b = b


class _DirectCycleB:
    def __init__(self, a: _DirectCycleA) -> None:
        self.a = a


class _RequestDependency:
    pass


class _RequestProviderConsumer:
    def __init__(self, dependency_provider: Provider[_RequestDependency]) -> None:
        self._dependency_provider = dependency_provider

    def get(self) -> _RequestDependency:
        return self._dependency_provider()


class _ScopedActionDependency:
    pass


class _RequestScopedProviderConsumer:
    def __init__(self, dependency_provider: Provider[_ScopedActionDependency]) -> None:
        self._dependency_provider = dependency_provider


class _VarArgDependency:
    pass


class _VarArgProviderConsumer:
    pass


def _build_vararg_provider_consumer(
    *providers: Provider[_VarArgDependency],
) -> _VarArgProviderConsumer:
    return _VarArgProviderConsumer()


class _VarKwProviderConsumer:
    pass


def _build_varkw_provider_consumer(
    **providers: Provider[_VarArgDependency],
) -> _VarKwProviderConsumer:
    return _VarKwProviderConsumer()


class _AsyncDependency:
    pass


class _AsyncConsumer:
    def __init__(self, dependency_provider: AsyncProvider[_AsyncDependency]) -> None:
        self._dependency_provider = dependency_provider

    async def get(self) -> _AsyncDependency:
        return await self._dependency_provider()


class _SyncProviderForAsyncDependency:
    def __init__(self, dependency_provider: Provider[_AsyncDependency]) -> None:
        self._dependency_provider = dependency_provider

    def get(self) -> _AsyncDependency:
        return self._dependency_provider()


class _InjectedConsumerDependency:
    pass


class _AutoregDependency:
    pass


class _AutoregConsumer:
    def __init__(self, dep_provider: Provider[_AutoregDependency]) -> None:
        self.dep_provider = dep_provider


@pytest.mark.asyncio
async def test_cycle_with_provider_breaks_assembly_cycle_and_resolves() -> None:
    container = Container()
    container.add(_CycleA)
    container.add(_CycleB)

    resolved = container.resolve(_CycleA)
    resolved_b = resolved.get_b()

    assert isinstance(resolved_b, _CycleB)
    assert resolved_b.a is resolved


def test_unbroken_direct_cycle_still_raises() -> None:
    container = Container()
    container.add(_DirectCycleA)
    container.add(_DirectCycleB)

    with pytest.raises(DIWireInvalidProviderSpecError, match="Circular dependency detected"):
        container.resolve(_DirectCycleA)


def test_provider_preserves_scoped_lifetime_within_same_scope() -> None:
    container = Container()
    container.add(
        _RequestDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _RequestProviderConsumer,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope() as request_scope:
        consumer = request_scope.resolve(_RequestProviderConsumer)
        first = consumer.get()
        second = consumer.get()

    assert first is second


def test_provider_preserves_transient_lifetime_within_same_scope() -> None:
    container = Container()
    container.add(
        _RequestDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add(
        _RequestProviderConsumer,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope() as request_scope:
        consumer = request_scope.resolve(_RequestProviderConsumer)
        first = consumer.get()
        second = consumer.get()

    assert first is not second


@pytest.mark.asyncio
async def test_async_provider_returns_awaitable_and_resolves_dependency() -> None:
    async def _build_dependency() -> _AsyncDependency:
        return _AsyncDependency()

    container = Container()
    container.add_factory(_build_dependency, provides=_AsyncDependency)
    container.add(_AsyncConsumer)

    consumer = container.resolve(_AsyncConsumer)
    dependency = await consumer.get()

    assert isinstance(dependency, _AsyncDependency)


def test_sync_provider_to_async_chain_raises_on_call() -> None:
    async def _build_dependency() -> _AsyncDependency:
        return _AsyncDependency()

    container = Container()
    container.add_factory(_build_dependency, provides=_AsyncDependency)
    container.add(_SyncProviderForAsyncDependency)

    consumer = container.resolve(_SyncProviderForAsyncDependency)

    with pytest.raises(
        DIWireAsyncDependencyInSyncContextError,
        match="requires asynchronous resolution",
    ):
        _ = consumer.get()


@pytest.mark.asyncio
async def test_direct_resolve_provider_and_async_provider_dependency_keys() -> None:
    container = Container()
    container.add(_InjectedConsumerDependency)
    resolver = container.compile()

    provider = resolver.resolve(Provider[_InjectedConsumerDependency])
    async_provider = resolver.resolve(AsyncProvider[_InjectedConsumerDependency])
    provider_from_async = await resolver.aresolve(Provider[_InjectedConsumerDependency])
    async_provider_from_async = await resolver.aresolve(AsyncProvider[_InjectedConsumerDependency])

    assert isinstance(provider(), _InjectedConsumerDependency)
    assert isinstance(await async_provider(), _InjectedConsumerDependency)
    assert isinstance(provider_from_async(), _InjectedConsumerDependency)
    assert isinstance(await async_provider_from_async(), _InjectedConsumerDependency)


@pytest.mark.asyncio
async def test_injected_wrapper_supports_provider_and_async_provider() -> None:
    container = Container()
    container.add(_InjectedConsumerDependency)

    @resolver_context.inject
    def _sync_handler(
        dependency_provider: Injected[Provider[_InjectedConsumerDependency]],
    ) -> _InjectedConsumerDependency:
        return dependency_provider()

    @resolver_context.inject
    async def _async_handler(
        dependency_provider: Injected[AsyncProvider[_InjectedConsumerDependency]],
    ) -> _InjectedConsumerDependency:
        return await dependency_provider()

    sync_handler = cast("Any", _sync_handler)
    async_handler = cast("Any", _async_handler)

    assert isinstance(sync_handler(), _InjectedConsumerDependency)
    assert isinstance(await async_handler(), _InjectedConsumerDependency)


T = TypeVar("T")


class _OpenProviderService(Generic[T]):
    pass


@dataclass
class _OpenProviderServiceImpl(_OpenProviderService[T]):
    type_arg: type[T]


def _build_open_provider_service(type_arg: type[T]) -> _OpenProviderService[T]:
    return _OpenProviderServiceImpl(type_arg=type_arg)


def test_provider_direct_resolve_supports_open_generic_dependencies() -> None:
    container = Container()
    container.add_factory(_build_open_provider_service, provides=_OpenProviderService)
    resolver = container.compile()

    provider = resolver.resolve(Provider[_OpenProviderService[int]])
    resolved = provider()

    assert isinstance(resolved, _OpenProviderServiceImpl)
    assert resolved.type_arg is int


def test_autoregistration_unwraps_provider_dependency() -> None:
    container = Container(
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE
    )
    container.add(_AutoregConsumer)

    assert container._providers_registrations.find_by_type(_AutoregDependency) is not None


def test_provider_rejects_deeper_scoped_dependency_during_planning() -> None:
    container = Container()
    container.add(
        _ScopedActionDependency,
        scope=Scope.ACTION,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _RequestScopedProviderConsumer,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(
        DIWireInvalidProviderSpecError,
        match="cannot bind deeper dependency",
    ):
        container.compile()


@pytest.mark.parametrize(
    "factory",
    [_build_vararg_provider_consumer, _build_varkw_provider_consumer],
)
def test_provider_rejects_star_parameter_shapes(factory: object) -> None:
    container = Container()
    container.add(_VarArgDependency)
    container.add_factory(
        cast("Any", factory),
        provides=(
            _VarArgProviderConsumer
            if factory is _build_vararg_provider_consumer
            else _VarKwProviderConsumer
        ),
    )

    with pytest.raises(
        DIWireInvalidProviderSpecError,
        match=r"star parameters \(\*args/\*\*kwargs\)",
    ):
        container.compile()
